'''
Implementation of nerf method
to calculate coords of a 4th
atom given the coords of the
first 3, the bond length (l),
bond angle (theta), and torsion
angle (chi)
Rohit Bhattacharya
rohit.bhattachar@gmail.com
'''

#ARTICLE: https://onlinelibrary.wiley.com/doi/10.1002/jcc.20237
# imports
import numpy as np

def nerf(a, b, c, l, theta, chi):
    '''
    Nerf method of finding 4th coord (d)
    in cartesian space
    Params:
    a, b, c : coords of 3 points
    l : bond length between c and d
    theta : bond angle between b, c, d (in degrees)
    chi : dihedral using a, b, c, d (in degrees)
    Returns:
    d : tuple of (x, y, z) in cartesian space
    '''

    # calculate unit vectors AB and BC
    ab_unit = (b-a)/np.linalg.norm(b-a)
    bc_unit = (c-b)/np.linalg.norm(c-b)

    # calculate unit normals n = AB x BC
    # and p = n x BC
    n_unit = np.cross(ab_unit, bc_unit)
    n_unit = n_unit/np.linalg.norm(n_unit)
    p_unit = np.cross(n_unit, bc_unit)

    # create rotation matrix [BC; p; n] (3x3)
    M = np.array([bc_unit, p_unit, n_unit]).T

    # convert degrees to radians
    theta = np.pi/180 * theta
    chi = np.pi/180 * chi

    # calculate coord pre rotation matrix
    d2 = [-l*np.cos(theta), l*np.sin(theta)*np.cos(chi), l*np.sin(theta)*np.sin(chi)]

    # calculate with rotation as our final output
    return c + np.dot(M, d2)


#####################################################################
# THESEUS IMPLEMENTATION (PROTEIN SUPERPOSITION) by LYS SANZ MORETA #
####################################################################
import os,sys
import time
from collections import defaultdict
import ntpath
import numpy as np
#Biopython
import Bio.PDB as PDB
from Bio.PDB.Polypeptide import is_aa
from Bio.SVDSuperimposer import SVDSuperimposer
#PYRO
import pyro
import pyro.distributions as dist
from pyro.infer.autoguide import AutoDelta,init_to_median
from pyro.infer import SVI,Trace_ELBO
from pyro.optim import PyroOptim
import matplotlib as mpl #If errors when running Pymol: conda install pyqt=5.6 downgrade
import matplotlib.pyplot as plt
import tqdm
#Early STOPPING
from ignite.handlers import EarlyStopping
from ignite.engine import Engine,Events
#TORCH: "Tensors"
import torch
import math
import pymol,pymol2

#Pre-set ups: Necessary for some torch errors
#torch.backends.cudnn.deterministic = True
PyroOptim.state_dict = lambda self: self.get_state()
mpl.use('agg') #Change matplotlib graphics backend to avoid conflict with Pymol
tqdm.monitor_interval = 0
class SVIEngine(Engine):
    def __init__(self, *args, step_args=None, **kwargs):
        self.svi = SVI(*args, **kwargs)
        self._step_args = step_args or {}
        super(SVIEngine, self).__init__(self._update)

    def _update(self, engine, batch):
        return -engine.svi.step(batch, **self._step_args)
class DataManagement():
    def __init__(self):
        super(DataManagement, self).__init__()
    def extract_coordinates_from_PDB(self,PDB_file,type):
        ''' Returns both the alpha carbon coordinates contained in the PDB file and the residues coordinates for the desired chains'''
        from Bio.PDB.PDBParser import PDBParser
        from Bio.PDB import MMCIFParser
        Name = ntpath.basename(PDB_file).split('.')[0]
        try:
            parser = PDB.PDBParser()
            structure = parser.get_structure('%s' % (Name), PDB_file)
        except:
            parser = MMCIFParser()
            structure = parser.get_structure('%s' % (Name), PDB_file)

        ############## Iterating over residues to extract all of them even if there is more than 1 chain
        if type=='models':
            CoordinatesPerModel = []
            for model in structure:
                model_coord =[]
                for chain in model:
                    for residue in chain:
                        if is_aa(residue.get_resname(), standard=True):
                                model_coord.append(residue['CA'].get_coord())
                CoordinatesPerModel.append(model_coord)

            return CoordinatesPerModel
        elif type=='chains':
            CoordinatesPerChain=[]
            for model in structure:
                for chain in model:
                    chain_coord = []
                    for residue in chain:
                        if is_aa(residue.get_resname(), standard=True):
                            chain_coord.append(residue['CA'].get_coord())
                    CoordinatesPerChain.append(chain_coord)
            return CoordinatesPerChain

        elif type =='all':
            alpha_carbon_coordinates = []
            for chain in structure.get_chains():
                for residue in chain:
                    if is_aa(residue.get_resname(), standard=True):
                        # try:
                        alpha_carbon_coordinates.append(residue['CA'].get_coord())
                    # except:
                    # pass
            return alpha_carbon_coordinates
    def average_structure(self,tuple_struct):
        average = sum(list(tuple_struct))/len(tuple_struct)
        return average
    def max_variance(self,structure):
        '''Calculates the maximum distance to the origin of the structure, this value will define the variance of the prior's distribution'''
        centered = self.center_torch(structure)
        mul = centered@torch.t(structure)
        max_var = torch.sqrt(torch.max(torch.diag(mul)))
        return max_var
    def center_numpy(self,Array):
        '''Centering to the origin the data'''
        mean = np.mean(Array,axis=0)
        centered_array = Array-mean
        return centered_array
    def center_torch(self,Array):
        '''Centering to the origin the data'''
        mean = torch.mean(Array, dim=0)
        centered_array = Array - mean
        return centered_array
    def pairwise_distances(self,X1,X2):
        '''Computes pairwise distances among coordinates'''
        import torch.nn.functional as F
        return F.pairwise_distance(X1,X2)
    def RMSD_biopython(self,x,y):
        """Kalbsch algorithm"""
        sup = SVDSuperimposer()
        sup.set(x, y)
        sup.run()
        rot, tran = sup.get_rotran()
        return rot
    def read_data(self,prot1,prot2,type='models',models =(0,1),RMSD=True):
        '''Reads different types of proteins and extracts the alpha carbons from the:
         a) models : NMR files contain several snapshots of the same protein
         b) chains: Proteins caan have dulicated chains in the same file
         c) all . We use for comparison a segment of each of the proteins
         '''

        if type == 'models': #we compare NMR models
            X1_coordinates = self.extract_coordinates_from_PDB('{}'.format(prot1),type)[models[0]]
            X2_coordinates = self.extract_coordinates_from_PDB('{}'.format(prot2),type)[models[1]]
        elif type == 'chains':#we compare chains in the same file against each other
            X1_coordinates = self.extract_coordinates_from_PDB('{}'.format(prot1),type)[models[0]][0:50]
            X2_coordinates = self.extract_coordinates_from_PDB('{}'.format(prot2),type)[models[1]][0:50]

        elif type == 'all': #we compare segments of proteins against each other
            X1_coordinates = self.extract_coordinates_from_PDB('{}'.format(prot1),type)[models[0]:models[1]]
            X2_coordinates = self.extract_coordinates_from_PDB('{}'.format(prot2),type)[models[0]:models[1]]
        #Apply RMSD to the protein that needs to be superimposed
        X1_Obs_Stacked = self.center_numpy(np.vstack(X1_coordinates))
        X2_Obs_Stacked = self.center_numpy(np.vstack(X2_coordinates))
        if RMSD:
            X2_Obs_Stacked = torch.from_numpy(np.dot(X2_Obs_Stacked,self.RMSD_biopython(X1_Obs_Stacked,X2_Obs_Stacked)))
            X1_Obs_Stacked = torch.from_numpy(X1_Obs_Stacked)
        else:
            X1_Obs_Stacked = torch.from_numpy(X1_Obs_Stacked)
            X2_Obs_Stacked = torch.from_numpy(X2_Obs_Stacked)

        data_obs = (X1_Obs_Stacked,X2_Obs_Stacked)

        # ###PLOT INPUT DATA################
        x= self.center_numpy(np.vstack(X1_coordinates))[:, 0]
        y= self.center_numpy(np.vstack(X1_coordinates))[:, 1]
        z= self.center_numpy(np.vstack(X1_coordinates))[:, 2]
        fig = plt.figure(figsize=(18, 16), dpi=80)
        ax = fig.add_subplot(111, projection='3d')
        plt.plot(x, y, z)
        ax.plot(x, y,z ,c='b', label='data1',linewidth=3.0)
        #orange graph
        x2=self.center_numpy(np.vstack(X2_coordinates))[:, 0]
        y2=self.center_numpy(np.vstack(X2_coordinates))[:, 1]
        z2=self.center_numpy(np.vstack(X2_coordinates))[:, 2]
        ax.plot(x2, y2,z2, c='r', label='data2',linewidth=3.0)
        ax.legend()
        plt.savefig(r"Initial.png")
        plt.clf() #Clear the plot
        plt.close()

        return data_obs
    def write_PDB(self,initialPDB,Rotation,Translation,N):
        ''' Transform by rotating and translating the atom coordinates from the original PDB file and rewrite it '''
        from Bio.PDB.PDBParser import PDBParser
        from Bio.PDB import MMCIFParser,PDBIO
        Name = ntpath.basename(initialPDB).split('.')[0]

        try:
            parser = PDB.PDBParser()
            structure = parser.get_structure('%s' % (Name), initialPDB)
        except:
            parser = MMCIFParser()
            structure = parser.get_structure('%s' % (Name), initialPDB)

        for atom in structure.get_atoms():
            atom.transform(Rotation, Translation)
        io = PDBIO()
        io.set_structure(structure)
        io.save("{}_{}".format(N,ntpath.basename(initialPDB)))
    def write_ATOM_line(self,structure, file_name):
        """Write a completely new PDB file with the C_alpha trace coordinates extracted from the model. It adds an intermediate coordinate between the C_alpha atoms in order to be able
        to visualize a conected line in Pymol"""
        #Create an expanded array: Contains an extra row between each C_alpha atom, for the intermediate coordinate
        expanded_structure = np.ones(shape=(2*len(structure)-1,3))
        #Calculate the average coordinate between each C alpha atom
        averagearray = np.zeros(shape=(len(structure)-1,3))
        for index,row in enumerate(structure):
            if index != len(structure) and index != len(structure)-1:
                averagearray[int(index)] = (structure[int(index)]+structure[int(index)+1])/2
            else:
                pass
        #The even rows of the 'expanded structure' are simply the rows of the original structure
        #The odd rows are the intermediate coordinate
        expanded_structure[0::2] = structure
        expanded_structure[1::2] = averagearray
        structure = expanded_structure
        aa_name = "ALA"
        aa_type="CA"
        if os.path.isfile(file_name):
            os.remove(file_name)
            for i in range(len(structure)):
                with open(file_name,'a') as f:
                    f.write("ATOM{:7d} {}   {} A{:4d}{:12.3f}{:8.3f}{:8.3f}  0.00  0.00    X    \n".format(i, aa_type, aa_name, i, structure[i, 0], structure[i, 1], structure[i, 2]))

        else:
            for i in range(len(structure)):
                with open(file_name,'a') as f:
                    f.write("ATOM{:7d} {}   {} A{:4d}{:12.3f}{:8.3f}{:8.3f}  0.00  0.00    X    \n".format(i, aa_type, aa_name, i, structure[i, 0], structure[i, 1], structure[i, 2]))
    def Pymol(self,results_folder,pdb_files):
        '''Visualization program. One time visualization'''
        #multiple instances of pymol: https://pymolwiki.org/index.php/Launching_From_a_Script
        #LAUNCH PYMOL
        launch=False
        if launch:
            pymol.pymol_argv = ['pymol'] + sys.argv[1:]
            pymol.finish_launching(['pymol'])
        def colour_backbone(selection,color):
            """Colouring function"""
            pymol.cmd.show("ribbon", selection)
            pymol.cmd.color(color,selection)
        # Iteratively load Structures and apply the function
        colornames=['red','green','blue','orange','purple','yellow','black','aquamarine']
        for file,color in zip(pdb_files,colornames):
            sname = ntpath.basename(file)
            pymol.cmd.load(file, sname) #discrete 1 will create different sets of atoms for each model
            pymol.cmd.enable(sname)
            pymol.cmd.bg_color("white")
            pymol.cmd.extend("Colour_Backbone", colour_backbone)
            colour_backbone(sname,color)
        pymol.cmd.align(ntpath.basename(pdb_files[0]),ntpath.basename(pdb_files[1]))
        pymol.cmd.png("{}/Superposition_Pymol".format(results_folder))
    def Pymol_newinstance(self,cmd, results_folder,pdb_files):
        '''Visualization program for multiple instances'''
        cmd.set("max_threads",1)
        def colour_backbone(selection,color):
            """Colouring function"""
            cmd.show("ribbon", selection)
            cmd.color(color,selection)
        # Iteratively load Structures and apply the function
        colornames=['red','green','blue','orange','purple','yellow','black','aquamarine']
        for file,color in zip(pdb_files,colornames):
            sname = ntpath.basename(file)
            cmd.load(file, sname) #discrete 1 will create different sets of atoms for each model
            cmd.enable(sname)
            cmd.bg_color("white")
            cmd.extend("Colour_Backbone", colour_backbone)
            colour_backbone(sname,color)
        cmd.align(ntpath.basename(pdb_files[0]),ntpath.basename(pdb_files[1]))
        #print("Saving to {}".format(results_folder))
        cmd.png("{}/Superposition_Pymol".format(results_folder),quiet=0)

class SuperpositionModel():
    def __init__(self):
        super(SuperpositionModel, self).__init__()
    def sample_R(self,ri_vec):
        """Inputs a sample of unit quaternion and transforms it into a rotation matrix"""
        theta1 = 2 * math.pi * ri_vec[1]
        theta2 = 2 * math.pi * ri_vec[2]

        r1 = torch.sqrt(1 - ri_vec[0])
        r2 = torch.sqrt(ri_vec[0])

        qw = r2 * torch.cos(theta2)
        qx = r1 * torch.sin(theta1)
        qy = r1 * torch.cos(theta1)
        qz = r2 * torch.sin(theta2)

        R= torch.eye(3,3)
        # Filling the rotation matrix
        # Row one
        R[0, 0]= qw**2 + qx**2 - qy**2 - qz**2
        R[0, 1]= 2*(qx*qy - qw*qz)
        R[0, 2]= 2*(qx*qz + qw*qy)

        # Row two
        R[1, 0]= 2*(qx*qy + qw*qz)
        R[1, 1]= qw**2 - qx**2 + qy**2 - qz**2
        R[1, 2]= 2*(qy*qz - qw*qx)

        # Row three
        R[2, 0]= 2*(qx*qz - qw*qy)
        R[2, 1]= 2*(qy*qz + qw*qx)
        R[2, 2]= qw**2 - qx**2 - qy**2 + qz**2
        return R
    def model(self,data):
        """Superposition of 2 proteins onto each other via rotation and translation"""
        max_var,data1, data2 = data
        ### 1. prior over mean M
        M = pyro.sample("M", dist.StudentT(1,0, 3).expand_by([data1.size(0),data1.size(1)]).to_event(2))
        ### 2. Prior over variances for the normal distribution
        U = pyro.sample("U", dist.HalfNormal(1).expand_by([data1.size(0)]).to_event(1))
        U =  U.reshape(data1.size(0),1).repeat(1,3).view(-1)  #Triplicate the rows for the subsequent mean calculation
        ## 3. prior over translations T_i: Sample translations for each of the x,y,z coordinates
        T2 = pyro.sample("T2", dist.Normal(0, 1).expand_by([3]).to_event(1))
        ## 4. prior over rotations R_i
        ri_vec = pyro.sample("ri_vec",dist.Uniform(0, 1).expand_by([3]).to_event(1))  # Uniform distribution
        R = self.sample_R(ri_vec)
        M_T1 = M
        M_R2_T2 = M @ R + T2
        # 5. Likelihood
        with pyro.plate("plate_univariate", data1.size(0)*data1.size(1),dim=-1):
            # pyro.sample("X1", dist.StudentT(1,M_T1.view(-1), U),obs=data1.view(-1))
            # pyro.sample("X2", dist.StudentT(1,M_R2_T2.view(-1), U), obs=data2.view(-1))
            pyro.sample("X1", dist.StudentT(1, M_T1.contiguous().view(-1), U), obs=data1.contiguous().view(-1))
            pyro.sample("X2", dist.StudentT(1, M_R2_T2.contiguous().view(-1), U), obs=data2.contiguous().view(-1))
    def run(self,data_obs,average,name,results_folder):
        #INITIALIZING PRIOR :
        def init_prior(site):
            if site["name"] == "ri_vec":
                return torch.tensor([0.9,0.1,0.9])
            elif site["name"] == "M":
                return average
            else:
                return init_to_median(site)
        #GUIDE
        global_guide = AutoDelta(self.model,init_loc_fn=init_prior)
        #OPTIMIZER
        optim =pyro.optim.AdagradRMSProp(dict())
        #ELBO
        elbo = Trace_ELBO()
        #STOCHASTIC VARIATIONAL INFERENCE
        svi_engine = SVIEngine(model =self.model,guide= global_guide,optim=optim,loss=elbo)
        pbar = tqdm.tqdm()
        loss_list = []

        #ERROR LOSS
        @svi_engine.on(Events.EPOCH_COMPLETED)
        def update_progress(svi_engine):
            pbar.update(1)
            loss_list.append(-svi_engine.state.output)
            plt.plot(loss_list)
            plt.savefig(r"{}/ELBO_Loss.png".format(results_folder))
            plt.close()

        # Register hooks to monitor gradient norms.
        svi_engine.run([data_obs],max_epochs=1)
        gradient_norms = defaultdict(list)
        for name_i, value in pyro.get_param_store().named_parameters():
            value.register_hook(lambda g, name_i=name_i: gradient_norms[name_i].append(g.norm().item()))

        #Earlystop
        handler = EarlyStopping(patience=25, score_function=lambda engine: engine.state.output, trainer=svi_engine)
        #SVI
        svi_engine.add_event_handler(Events.EPOCH_COMPLETED, handler)
        start = time.time()
        svi_engine.run([data_obs],max_epochs=15000)
        stop=time.time()
        pbar.refresh()
        pbar.close()
        duration = stop-start
        ###PLOTTING THE GRADIENT
        plt.figure(figsize=(10, 4), dpi=100).set_facecolor('white')
        for name_i, grad_norms in gradient_norms.items():
            plt.plot(grad_norms, label=name_i)
        plt.xlabel('iters')
        plt.ylabel('gradient norm')
        plt.yscale('log')
        plt.legend(loc='best')
        plt.title('Gradient norms during SVI')
        plt.savefig("{}/Gradients".format(results_folder))
        plt.close()
        #### PARAMETERS ################
        max_var,data1,data2 = data_obs
        map_estimates = global_guide(data_obs)
        #Mean structure
        M = map_estimates["M"].detach().numpy()
        #Rotations
        ri_vec = map_estimates["ri_vec"].detach()
        R = self.sample_R(ri_vec)
        #Translation
        T2 = map_estimates["T2"].detach()
        #Observed
        X1 = data1.detach().numpy()  #X1
        X2 = np.dot(data2.detach().numpy() - T2.numpy(), np.transpose(R)) #(X2-T2)R-1

        #################PLOTS################################################
        fig = plt.figure(figsize=(18, 16), dpi=80)
        ax = fig.add_subplot(111, projection='3d')

        #blue graph
        x = X1[:,0]
        y = X1[:,1]
        z = X1[:,2]

        ax.plot(x, y,z ,c='b', label='X1',linewidth=3.0)

        #red graph
        x2 = X2[:,0]
        y2 = X2[:,1]
        z2 = X2[:,2]

        ax.plot(x2, y2,z2, c='r', label='X2',linewidth=3.0)

        ###green graph
        x3=M[:,0]
        y3=M[:,1]
        z3=M[:,2]
        ax.plot(x3, y3,z3, c='g', label='M',linewidth=3.0)
        ax.legend()
        plt.savefig(r"{}/Superposition_Result_Matplotlib_{}".format(results_folder,name))
        distances = data_management.pairwise_distances(torch.from_numpy(X1),torch.from_numpy(X2)).numpy()
        plt.clf()
        plt.plot(data_management.pairwise_distances(data1,data2).numpy(), linewidth = 8.0)
        plt.plot(data_management.pairwise_distances(torch.from_numpy(X1),torch.from_numpy(X2)).numpy(), linewidth=8.0)
        plt.ylabel('Pairwise distances',fontsize='46')
        plt.xlabel('Amino acid position',fontsize='46')
        plt.yticks(fontsize=30)
        plt.xticks(fontsize=30)
        plt.title('{}'.format(name.upper()),fontsize ='46')
        plt.gca().legend(('RMSD', 'Theseus-PP'),fontsize='40')
        plt.savefig(r"{}/Distance_Differences_{}".format(results_folder,name))
        plt.close()

        return T2.numpy(),R.numpy(),M, X1,X2,distances,(svi_engine.state.epoch,svi_engine.state.output),duration

data_management = DataManagement()
superposition_model = SuperpositionModel()
















'''
Script to calculate coords in a peptide backbone
using the nerf method, given a file containing
number of residues in the backbone and pairs
of phi psi angles
Rohit Bhattacharya
rohit.bhattachar@gmail.com
'''

# imports
from nerf import nerf
import numpy as np
import argparse


def parse_args():
    '''
    Parse arguments
    '''

    info = 'Calculate coords given torsions file'
    parser = argparse.ArgumentParser(description=info)

    parser.add_argument('-i', '--input',
                        type=str, required=True,
                        help='Torsions input file')

    args = parser.parse_args()
    return vars(args)


def place_residues(torsions):
    '''
    Do the actual placement of residues
    based on nerf method.
    '''

    # place the 0th carbonyl C at origin
    n_i_zero = np.array([0.0, 0.0, 0.0])

    # using CH1E-NH1 bond length = 1.458 ang
    # place C_alpha on pos x axis
    c_a_zero = np.array([1.458, 0.0, 0.0])

    # place C_carbonyl on xy plane using
    # CH1E-C bond length = 1.525 ang
    # C-CH1E-NH1 bond angle = 111.2 deg
    l = 1.525
    theta = 111.2 * np.pi/180  # in radians
    c_o_zero = c_a_zero + np.array([-l*np.cos(theta), l*np.sin(theta), 0])

    # start a listing of atoms starting with the zeroth
    # residue represented as a tuple of N, C_alpha, C_carbonyl
    atom_coords = [(n_i_zero, c_a_zero, c_o_zero)]

    # go through each set of torsion angles

    for i in range(1, len(torsions)):

        # calculate coords of all atoms of
        # the ith residue based on coords
        # from i-1th residue
        n_i, c_a, c_o = atom_coords[i-1]
        phi, psi, omega = torsions[i]
        n_i_plus_1 = nerf(n_i, c_a, c_o, l=1.329, theta=116.2, chi=psi)
        c_a_i_plus_1 = nerf(c_a, c_o, n_i_plus_1,l=1.458, theta=121.7, chi=omega)
        c_o_i_plus_1 = nerf(c_o, n_i_plus_1, c_a_i_plus_1,l=1.525, theta=111.2, chi=phi)
        atom_coords.append((n_i_plus_1, c_a_i_plus_1, c_o_i_plus_1))

    # return the calculated coords
    return atom_coords

def format_coodinate(coordinate):
    n_integers = len(str(int(coordinate)))
    n_decimals = 3 #rounded up to 3


def write_to_pdb(residue_coords, filename = "torsions.pdb"):
    '''
    Write out calculated coords
    to PDB
    '''

    atom_names = ['N', 'CA', 'C']*(len(residue_coords))
    atoms_short_name = ["N","C","C"]*(len(residue_coords))
    out_file = open(filename, 'w')
    atom_coords = []
    for res_tuple in residue_coords:
        for atom_coord in res_tuple:
            atom_coords.append([round(c, 3) for c in atom_coord])

    res_number = 0
    for i, coord in enumerate(atom_coords):

        if i % 3 == 0:
            res_number += 1
        line = [" "]*80
        line[0:6] = ["A","T","O","M"," "," "]
        line[6:11] = [" "]*(5-len(list(str(i+1)))) + list(str(i+1)) #atom serial number,start counting at 1
        line[12:16] = [" "]*(4-len(list(atom_names[i]))) + list(atom_names[i]) #atom full name
        line[17:20] = list("ALA") #residue name
        line[21] = "A" #chain
        line[22:26] = [" "]*(4-len(list(str(res_number)))) +list(str(res_number)) #residue sequence number
        line[26] = " " #insertion code
        line[30:38] = [" "]*(8-len(list("{0:.3f}".format(coord[0])))) + list("{0:.3f}".format(coord[0])) #x
        line[38:46] = [" "]*(8-len(list("{0:.3f}".format(coord[1])))) + list("{0:.3f}".format(coord[1])) #y
        line[46:54] = [" "]*(8-len(list("{0:.3f}".format(coord[2])))) + list("{0:.3f}".format(coord[2])) #z
        # line[30:38] = [" "] * (8 - len(list(str(coord[0])))) + list(str(coord[0]))  # x #works either way
        # line[38:46] = [" "] * (8 - len(list(str(coord[1])))) + list(str(coord[1]))  # y
        # line[46:54] = [" "] * (8 - len(list(str(coord[2])))) + list(str(coord[2]))  # z
        line[54:60] =  [" "]*(5-len(list(str(1.02)))) +list(str(1.02)) #occupancy
        line[60:66] = [" "]*(5-len(list(str(11.92)))) + list(str(11.92)) #bfactor?
        #line[72:76] = list(str("A1")) #segid
        line[77:78] = list(atoms_short_name[i])
        out_file.write("{}\n".format("".join(line)))
    out_file.close()


def main():

    opts = parse_args()

    # make sure all the inputs
    # look okay
    try:
        in_file = open(opts['input'])
        n_residues = int(in_file.readline().strip())

    except IOError:
        print('Input file', opts['input'], 'does not exist')

    except ValueError:
        print('Num residues must be an integer')

    torsions = []
    for line in in_file:
        line = line.strip().split(',')

        # must be a pair of phi psi angles
        assert len(line) == 2, 'Must have a pair of comma separated phi psi angles'
        try:
            phi = float(line[0])
            psi = float(line[1])
            omega = 180
        except ValueError:
            print ('Angles must be valid float')
        torsions.append((phi, psi, omega))

    # finally make sure we have same number of phi psi
    # pairs as number of residues
    try:
        assert len(torsions) == n_residues, 'Must have same number of torsions as residues'

    except AssertionError as e:
        print(e)

    # actually calculate the coordinates
    calculated_coords = place_residues(torsions)
    write_to_pdb(calculated_coords)

if __name__ == '__main__':
    main()


"""
2021: aleatoryscience
Lys Sanz Moreta
Draupnir : GP prior VAE for Ancestral Sequence Resurrection
=======================
"""
import torch
from collections import defaultdict,namedtuple

import Draupnir_utils
from Draupnir_models_utils import *
import pyro
import pyro.distributions as dist
from pyro.util import ignore_jit_warnings
try:
    from torch_geometric.data import Data
except:
    pass
SamplingOutput = namedtuple("SamplingOutput",["aa_sequences","latent_space","logits","phis","psis","mean_phi","mean_psi","kappa_phi","kappa_psi"])
class DRAUPNIRModelClass(nn.Module):
    def __init__(self, ModelLoad):
        super(DRAUPNIRModelClass, self).__init__()
        self.num_epochs = ModelLoad.args.num_epochs
        self.gru_hidden_dim = ModelLoad.gru_hidden_dim
        self.embedding_dim = ModelLoad.args.embedding_dim #blosum embedding dim
        self.position_embedding_dim= ModelLoad.args.position_embedding_dim
        self.pretrained_params = ModelLoad.pretrained_params
        self.z_dim = ModelLoad.z_dim
        self.leaves_nodes = ModelLoad.leaves_nodes
        self.n_tree_levels = ModelLoad.n_tree_levels
        self.max_seq_len = ModelLoad.max_seq_len
        self.aa_prob = ModelLoad.build_config.aa_prob
        self.edge_info = ModelLoad.graph_coo #for the gnn + gru hybrid model
        self.nodes_representations_array = ModelLoad.nodes_representations_array
        self.dgl_graph = ModelLoad.dgl_graph
        self.children_dict = ModelLoad.children_dict
        self.closest_leaves_dict = ModelLoad.closest_leaves_dict
        self.descendants_dict = ModelLoad.descendants_dict
        self.clades_dict_all = ModelLoad.clades_dict_all
        self.max_indel_size = ModelLoad.args.max_indel_size
        self.rnn_input_size = self.z_dim
        self.use_attention = False
        self.batch_first = True
        self.leaves_testing = ModelLoad.leaves_testing
        self.batch_by_clade = ModelLoad.args.batch_by_clade
        self.device = ModelLoad.device
        self.kappa_addition = ModelLoad.args.kappa_addition
        self.aa_frequencies = ModelLoad.aa_frequencies
        self.blosum = ModelLoad.blosum
        self.blosum_max = ModelLoad.blosum_max
        self.blosum_weighted = ModelLoad.blosum_weighted
        self.variable_score = ModelLoad.variable_score
        self.internal_nodes = ModelLoad.internal_nodes
        self.batch_size = ModelLoad.build_config.batch_size
        self.plate_size = ModelLoad.build_config.plate_subsample_size
        if self.batch_size > 1 : # for normal batching and  batch by clade
            self.n_leaves = self.batch_size
            self.n_internal = len(self.internal_nodes)
            self.n_all = self.n_leaves + self.n_internal
        else:
            self.n_leaves = len(self.leaves_nodes)
            self.n_internal = len(self.internal_nodes)
            self.n_all = self.n_leaves + self.n_internal
        self.num_layers = 1 #TODO: Remove
        self.h_0_MODEL = nn.Parameter(torch.rand(self.gru_hidden_dim), requires_grad=True).to(self.device)
        self.decoder = RNNDecoder_FCL(self.max_seq_len, self.aa_prob,self.gru_hidden_dim, self.z_dim, self.rnn_input_size,self.kappa_addition,self.batch_first)
        if ModelLoad.args.use_cuda:
            self.cuda()
    @abstractmethod
    def guide(self,family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum):
        raise NotImplementedError

    @abstractmethod
    def model(self, family_data,patristic_matrix,cladistic_matrix,clade_blosum):
        raise NotImplementedError

    @abstractmethod
    def sample(self, map_estimates, n_samples, family_data_test, patristic_matrix,cladistic_matrix,use_test=True):
        raise NotImplementedError
    @abstractmethod
    def get_class(self):
        full_name = self.__class__
        name = str(full_name).split(".")[-1].replace("'>","")
        return name

class DRAUPNIRModel_VAE(DRAUPNIRModelClass):
    "Blosum weighted average embedding. Training on leaves, testing on internal nodes, no batching"
    def __init__(self,ModelLoad):
        DRAUPNIRModelClass.__init__(self,ModelLoad)
        self.rnn_input_size = self.z_dim + self.aa_prob
        self.decoder = RNNDecoder_Tiling(self.max_seq_len, self.aa_prob, self.gru_hidden_dim, self.z_dim, self.rnn_input_size,self.kappa_addition,self.num_layers,self.pretrained_params)
        self.embed = EmbedComplex(self.aa_prob,self.embedding_dim, self.pretrained_params)

    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum = None):
        aminoacid_sequences = family_data[:, 2:, 0]
        batch_nodes = family_data[:, 0, 1]
        batch_indexes = (patristic_matrix_sorted[1:, 0][..., None] == batch_nodes).any(-1)
        # Highlight: Register GRU module
        pyro.module("embeddings",self.embed)
        pyro.module("decoder_attention", self.decoder_attention)
        pyro.module("decoder", self.decoder)
        # Highlight: No prior over the latent space
        #latent_space = self.gp_prior(patristic_matrix_sorted)
        covariance = torch.eye(self.n_leaves).repeat(self.z_dim,1,1)
        mean = torch.zeros(self.z_dim,self.n_leaves)
        latent_space = pyro.sample("latent_z",dist.MultivariateNormal(mean, covariance)).T
        # Highlight: MAP the latent space to logits using GRU
        latent_space = latent_space.repeat(1,self.max_seq_len).reshape(latent_space.shape[0],self.max_seq_len,self.z_dim) #[n_nodes,max_seq,z_dim]

        blosum = self.blosum_weighted.repeat(latent_space.shape[0],1).reshape(latent_space.shape[0],self.max_seq_len,self.aa_prob) #[n_nodes,max_seq,21]
        blosum = self.embed(blosum)
        latent_space = torch.cat((latent_space,blosum),dim=2) #[n_nodes,max_seq_len,z_dim + 21]
        decoder_hidden = self.h_0_MODEL.expand(self.decoder.num_layers * 2, latent_space.shape[0],
                                               self.gru_hidden_dim).contiguous()  # bidirectional

        with pyro.plate("plate_len", aminoacid_sequences.shape[1], dim=-1):
            with pyro.plate("plate_seq", aminoacid_sequences.shape[0], dim=-2, subsample_size=self.plate_size) as indx:
                logits = self.decoder.forward(
                    input=latent_space[indx],
                    hidden=decoder_hidden[:, indx])
                pyro.sample("aa_sequences", dist.Categorical(logits=logits),
                            obs=aminoacid_sequences[indx])  # aa_seq = [n_nodes,max_seq_len]

    def sample(self, map_estimates, n_samples, family_data_test, patristic_matrix,cladistic_matrix,use_argmax=False,use_test=True,use_test2=False):

        latent_space = map_estimates["latent_z"].T
        assert latent_space.shape == (self.n_leaves, self.z_dim)
        n_nodes = self.n_leaves

        decoder_hidden = self.h_0_MODEL.expand(self.decoder.num_layers * 2, latent_space.shape[0],self.gru_hidden_dim).contiguous()  # Not bidirectional
        latent_space_ = latent_space.repeat(1, self.max_seq_len).reshape(n_nodes,self.max_seq_len, self.z_dim)
        blosum = self.blosum_weighted.repeat(latent_space_.shape[0], 1).reshape(latent_space_.shape[0], self.max_seq_len,self.aa_prob)  # [n_nodes,max_seq,21]
        blosum = self.embed(blosum)
        latent_space_ = torch.cat((latent_space_, blosum), dim=2)  # [n_nodes,max_seq_len,z_dim + 21]

        with pyro.plate("plate_len",self.max_seq_len, dim=-1), pyro.plate("plate_seq",n_nodes,dim=-2,subsample_size=n_nodes):
            logits = self.decoder.forward(
                input=latent_space_,
                hidden=decoder_hidden)
            if use_argmax:
                #Pick the sequence with the highest likelihood, now n_samples, n_samples = 1
                aa_sequences = torch.argmax(logits,dim=2).unsqueeze(0) #I add one dimension at the beginning to resemble 1 sample and not have to change all the plotting code
            else:
                aa_sequences = dist.Categorical(logits=logits).sample([n_samples])
        sampling_out = SamplingOutput(aa_sequences=aa_sequences.detach(),
                                      latent_space=latent_space.detach(),
                                      logits=logits.detach(),
                                      phis=None,
                                      psis=None,
                                      mean_phi=None,
                                      mean_psi=None,
                                      kappa_phi=None,
                                      kappa_psi=None)

        return sampling_out



from ete3 import Tree, faces, AttrFace, TreeStyle, NodeStyle, Tree, TextFace
# from PyQt4 import QtCore
# from PyQt4.QtGui import QGraphicsRectItem, QGraphicsSimpleTextItem, \
#     QGraphicsEllipseItem, QColor, QPen, QBrush
import matplotlib.pyplot as plt
import matplotlib
import pickle, os

def renaming():
    "Rename the internal nodes, unless the given newick file already has the names on it"
    # Rename the internal nodes
    leafs_names = tree.get_leaf_names()
    edge = len(leafs_names)
    internal_nodes_names = []
    for node in tree.traverse():  # levelorder (nodes are visited in zig zag order from root to leaves)
        if not node.is_leaf():
            node.name = "A%d" % edge
            internal_nodes_names.append(node.name)
            edge += 1


def renaming_simulations(with_indexes=False):
    "Rename the internal nodes to I + number, in simulations the leaves have the prefix A. With indexes shows the names used when transferring to an array"
    print("Renaming tree from simulations")
    leafs_names = tree.get_leaf_names()
    edge = len(leafs_names)
    internal_nodes_names = []
    if with_indexes:
        for idx, node in enumerate(
                tree.traverse()):  # levelorder (nodes are visited in zig zag order from root to leaves)
            if not node.is_leaf():
                node.name = "I" + node.name + "/{}".format(idx)
                internal_nodes_names.append(node.name)
                edge += 1
            else:
                node.name = node.name + "/{}".format(idx)
                # edge += 1

    else:
        for node in tree.traverse():  # levelorder (nodes are visited in zig zag order from root to leaves)
            if not node.is_leaf():
                node.name = "I" + node.name
                internal_nodes_names.append(node.name)
                edge += 1

def my_layout(node):
    "Adds the internal nodes names"
    if node.is_leaf():
        # If terminal node, draws its name
        name_face = AttrFace("name",fsize=8,fgcolor="black")
    else:
        # If internal node, draws label with smaller font size
        name_face = AttrFace("name", fsize=8,fgcolor="black")
    # Adds the name face to the image at the preferred position
    faces.add_face_to_node(name_face, node, column=0, position="branch-right")
def colour_tree_by_clades(clades_dict, tree, rename_internal_nodes=True):
    color_map21 = matplotlib.colors.ListedColormap(
        ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
         "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
         "black"])

    color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'gist_rainbow'][0]
    clrs = plt.get_cmap(color_map_name, len(clades_dict))
    clrs = [matplotlib.colors.rgb2hex(clrs(i)) for i in range(clrs.N)]
    node_styles = []
    for color in clrs:
        ns = NodeStyle()
        ns["fgcolor"] = color #tips color
        ns["size"] = 6
        ns["vt_line_width"] = 1
        ns["hz_line_width"] = 1
        node_styles.append(ns)
    for node_style,(clade, nodes) in zip(node_styles,clades_dict.items()):
        clade_ancestor = tree.get_common_ancestor(nodes["leaves"])
        clade_ancestor.set_style(node_style)
        for node in clade_ancestor.get_descendants():
            node.set_style(node_style)

    if rename_internal_nodes:
        if name.startswith("simulations"): renaming_simulations(with_indexes=True)
        else:renaming()
    ts = TreeStyle()
    # Do not add leaf names automatically
    ts.show_leaf_name = False
    # Use my custom layout
    #ts.layout_fn = my_layout
    ts.show_branch_length = False
    ts.show_scale = False
    #new stuff
    # ts.legend= False
    # ts.force_topology = True
    # ts.show_border = True
    # ts.scale_length = False
    try:
        tree.render("Tree_Pictures/return_{}_colored_by_clades_stripped.png".format(name), w=1000, units="mm", tree_style=ts)
    except:
        tree.render("Tree_Pictures/return_{}_colored_by_clades_stripped.png".format(name), w=1000, units="mm")




if __name__ == "__main__":
    #t, ts = get_example_tree()
    script_dir = os.path.dirname(os.path.abspath(__file__))

    datasets = {0:["benchmark_randall", None, None, None],  #the tree is inferred
                1:["benchmark_randall_original",None, None,None],  #uses the original tree but changes the naming of the nodes (because the original tree was not rooted)
                2:["benchmark_randall_original_naming",None,None,None],  #uses the original tree and it's original node naming
                3:["SH3_pf00018_larger_than_30aa",None,None,None],  #SRC kinases domain SH3 ---> Leaves and angles testing
                4:["simulations_blactamase_1",1,"BLactamase","BetaLactamase_seq"],  #EvolveAGene4 Betalactamase simulation # 32 leaves
                5:["simulations_src_sh3_1",1, "SRC_simulations","SRC_SH3"],  #EvolveAGene4 SRC SH3 domain simulation 1 #100 leaves
                6:["simulations_src_sh3_2",2, "SRC_simulations","SRC_SH3"],  #EvolveAGene4 SRC SH3 domain simulation 2 #800 leaves
                7: ["simulations_src_sh3_3", 3, "SRC_simulations", "SRC_SH3"],# EvolveAGene4 SRC SH3 domain simulation 2 #200 leaves
                8: ["simulations_sirtuins_1",1, "Sirtuin_simulations", "Sirtuin_seq"],  # EvolveAGene4 Sirtuin simulation #150 leaves
                9: ["simulations_calcitonin_1", 1, "Calcitonin_simulations", "Calcitonin_seq"],# EvolveAGene4 Calcitonin simulation #50 leaves
                10: ["simulations_mciz_1",1, "Mciz_simulations","Mciz_seq"],  # EvolveAGene4 MciZ simulation # 1600 leaves
                11:["Douglas_SRC",None,None,None],  #Douglas's Full SRC Kinases #Highlight: The tree is not similar to the one in the paper, therefore the sequences where splitted in subtrees according to the ancetral sequences in the paper
                12:["ANC_A1_subtree",None,None,None],  #highlight: 3D structure not available #TODO: only working at dragon server
                13:["ANC_A2_subtree",None,None,None],  #highlight: 3D structure not available
                14:["ANC_AS_subtree",None,None,None],
                15:["ANC_S1_subtree",None,None,None],  #highlight: 3D structure not available
                16:["Coral_Faviina",None,None,None],  #Faviina clade from coral sequences
                17:["Coral_all",None,None,None],  # All Coral sequences (includes Faviina clade and additional sequences)
                18:["Cnidarian",None,None,None],  # All Coral sequences plus other fluorescent cnidarians #Highlight: The tree is too different to certainly locate the all-coral / all-fav ancestors
                19:["PKinase_PF07714",None,None,None],
                20:["simulations_CNP_1", 1, "CNP_simulations", "CNP_seq"],  # EvolveAGene4 CNP simulation # 1000 leaves
                21:["simulations_insulin_1", 1, "Insulin_simulations", "Insulin_seq"],  # EvolveAGene4 Insulin simulation #50 leaves
                22:["PF01038_msa",None,None,None],
                }
    name,dataset_number,simulation_folder,root_sequence_name = datasets[9]
    clades_dict_all = pickle.load(open('{}/Mixed_info_Folder/{}_Clades_dict_all.p'.format(script_dir, name), "rb"))
    if name.startswith("simulations"):
        tree_file = "Datasets_Simulations/{}/Dataset{}/{}_True_Rooted_tree_node_labels.tre".format(simulation_folder,dataset_number,root_sequence_name)
    elif name == "benchmark_randall_original_naming":
        tree_file = "AncestralResurrectionStandardDataset/RandallBenchmarkTree_OriginalNaming.tree"
    elif name.endswith("_subtree"):
        tree_file = "/home/lys/Dropbox/PhD/DRAUPNIR/Douglas_SRC_Dataset/{}/{}_ALIGNED.mafft.treefile".format(name,name.replace("_subtree",""))
    else:
        tree_file = "Mixed_info_Folder/{}.mafft.treefile".format(name)
    tree = Tree(tree_file, format=1, quoted_node_names=True)
    colour_tree_by_clades(clades_dict_all,tree)



import os, itertools,sys
from collections import defaultdict
import torch
import numpy as np
sys.path.append("./draupnir/draupnir")
import Draupnir_utils as DraupnirUtils
import Draupnir_models_utils as DraupnirModelUtils
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.pylab as pl
import matplotlib
from mpl_toolkits.mplot3d import Axes3D
from Bio import SeqRecord, SeqIO
from Bio.Seq import Seq
import seaborn as sns
from sklearn.decomposition import PCA
from Bio.SeqUtils import seq3
from sklearn.manifold import TSNE
import statistics
import umap
from scipy import stats
def plot_ELBO(train_elbo,results_dict):
    """Plots the model's error loss
    :param list train_elbo: list of accumulated error losses
    :param str results_dict: path to results directory
    """
    train_elbo = np.array(train_elbo)
    list_epochs = list(range(0,len(train_elbo)))
    if np.isnan(train_elbo).any():
        print("Error loss contains nan")
        pass
    else:
        plt.plot(list_epochs,train_elbo,color="dodgerblue")
        plt.xlabel("List of epochs")
        plt.ylabel("-ELBO")
        plt.title("Training Error Loss (min -ELBO, min KL)")
        plt.savefig("{}/ELBO_error.png".format(results_dict))
        plt.close()
def plot_entropy(train_entropy,results_dict):
    """Plots the model's entropy
    :param list train_entropy: list of accumulated entropies
    :param str results_dict: path to results directory
    """
    train_entropy = np.array(train_entropy)
    list_epochs = list(range(0, len(train_entropy)))
    if np.isnan(train_entropy).any():
        print("Entropy contains nan")
        pass
    else:
        plt.plot(list_epochs, train_entropy, color="tomato")
        plt.title("Shanon Entropy Convergence")
        plt.savefig("{}/Entropy_convergence.png".format(results_dict))
        plt.close()
def plot_z(latent_space, children_dict, results_dir):
    """Plots the latent space vectors
    :param tensor latent_space
    :param dict children_dict
    :param str results_dir
    """

    fig, ax = plt.subplots(figsize=(50, 30))  # 18 for 1 col
    colors = plt.get_cmap('cool', len(children_dict.keys()))
    markers = itertools.cycle(('s', '+', 'd', 'o', '*'))
    for idx, (ancestor, children) in enumerate(children_dict.items()):
        if not pd.isnull(ancestor):
            sequences = [ancestor] + children
            indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
            sequences_latent_space = latent_space[indexes,1:]
            for seq in range(len(sequences)):
                ax.plot(sequences_latent_space[seq,1:],color=colors(idx),marker=next(markers))
    plt.ylabel(r"Latent space ($\mathcal{Z}$) vector",fontsize=40)
    plt.xlabel(r"$\mathcal{Z}$ dimensions",fontsize=40)
    plt.title(r"Latent space ($\mathcal{Z}$) vector coloured by ancestor and respective children nodes",fontsize=40)
    plt.savefig("{}/z_vector_plot".format(results_dir))
def plot_angles(samples_out,dataset_test,results_dir,additional_load,additional_info,n_samples,test_ordered_nodes):
    """Plot Ramachandran plots of the predicted angles, both per individual amino acid and all amino acids combined
    :param namedtuple samples_out: contains the output from Draupnir.samples
    :param tensor dataset_test
    :param str results_dir
    :param namedtuple additional_load
    :param namedtuple additional_info
    :param int n_samples
    :param tensor test_ordered_nodes: tensor containing the associated nodes indexes"""

    phi_angles = samples_out.phis
    psi_angles = samples_out.psis
    phi_mean = samples_out.mean_phi
    psi_mean = samples_out.mean_psi
    phi_kappa = samples_out.kappa_phi
    psi_kappa = samples_out.kappa_psi
    #phi_angles = dataset_test[:,2:,1].repeat(n_samples,1,1)
    #psi_angles = dataset_test[:, 2:, 2].repeat(n_samples, 1, 1)
    Dataset_mask = DraupnirModelUtils.masking(dataset_test[:,2:,0]) #where there is a GAP, aka 0, put a 0 in the mask
    DraupnirUtils.ramachandran_plot_sampled(phi_angles, psi_angles,"{}/Angles_Predictions.png".format(results_dir), r"Sampled angles (φ,ψ); {}".format(additional_load.full_name))
    DraupnirUtils.ramachandran_plot_sampled(phi_mean, psi_mean, "{}/Angles_MEANS_Predictions.png".format(results_dir),
                                            r"Sampled angles's means (φ,ψ); {}".format(additional_load.full_name))
    DraupnirUtils.ramachandran_plot_sampled(phi_kappa, psi_kappa, "{}/Angles_KAPPAS_Predictions.png".format(results_dir),
                                            r"Sampled angles's kappas (φ,ψ); {}".format(additional_load.full_name),plot_kappas=True)

    phi_angles_list = []
    psi_angles_list = []
    phi_mean_list= []
    psi_mean_list=[]
    phi_kappa_list = []
    psi_kappa_list = []
    for i in range(n_samples): #TODO: Merge with angles per aa
        for row_mask,row_phi,row_psi in zip(Dataset_mask,phi_angles[i],psi_angles[i]):
            phi_angles_list.append(row_phi[row_mask !=0].tolist())
            psi_angles_list.append(row_psi[row_mask != 0].tolist())

    for row_mask,row_phi_mean, row_psi_mean, row_phi_kappa, row_psi_kappa in zip(Dataset_mask, phi_mean, psi_mean, phi_kappa, psi_kappa):
            phi_mean_list.append(row_phi_mean[row_mask != 0].tolist())
            psi_mean_list.append(row_psi_mean[row_mask != 0].tolist())
            phi_kappa_list.append(row_phi_kappa[row_mask != 0].tolist())
            psi_kappa_list.append(row_psi_kappa[row_mask != 0].tolist())

    DraupnirUtils.ramachandran_plot_sampled(sum(phi_angles_list,[]),sum(psi_angles_list,[]),"{}/Angles_Predictions_NoGAPS".format(results_dir),r"Sampled angles (φ,ψ) (without GAPS); {}".format(additional_load.full_name))
    DraupnirUtils.ramachandran_plot_sampled(sum(phi_mean_list,[]),sum(psi_mean_list,[]),"{}/Angles_MEANS_Predictions_NoGAPS".format(results_dir),r"Sampled angles's means (φ,ψ) (without GAPS); {}".format(additional_load.full_name))
    DraupnirUtils.ramachandran_plot_sampled(sum(phi_kappa_list,[]),sum(psi_kappa_list,[]),"{}/Angles_KAPPAS_Predictions_NoGAPS".format(results_dir),r"Sampled angles's kappas (φ,ψ) (without GAPS); {}".format(additional_load.full_name),plot_kappas=True)
def plot_angles_per_aa(samples_out,dataset_test,results_dir,build_config,additional_load,additional_info,n_samples,test_ordered_nodes):
    """Plot Ramachandran plots of the predicted angles, both per individual amino acid and all amino acids combined
    :param namedtuple samples_out: contains the output from Draupnir.samples
    :param tensor dataset_test
    :param str results_dir
    :param namedtuple additional_load
    :param namedtuple additional_info
    :param int n_samples
    :param tensor test_ordered_nodes: tensor containing the associated nodes indexes"""

    phi_angles = samples_out.phis
    psi_angles = samples_out.psis
    amino_acid_dict = DraupnirUtils.aminoacid_names_dict(build_config.aa_prob)
    amino_acids_data = dataset_test[:,2:,0]
    for aa_name, aa_number in amino_acid_dict.items():
        if aa_name == "-": continue
        else:
            phi_angles_list = []
            psi_angles_list = []
            for i in range(n_samples):
                for row_data,row_phi,row_psi in zip(amino_acids_data,phi_angles[i],psi_angles[i]):
                    phi_angles_list.append(row_phi[(row_data !=0)&(row_data == aa_number)].tolist())
                    psi_angles_list.append(row_psi[(row_data !=0)&(row_data == aa_number)].tolist())

            DraupnirUtils.ramachandran_plot_sampled(sum(phi_angles_list,[]),sum(psi_angles_list,[]),"{}/Angles_plots_per_aa/Angles_Predictions_{}".format(results_dir,aa_name),r"Sampled angles (φ,ψ) {} (without GAPS); {}".format(seq3(aa_name),additional_load.full_name))
    # DraupnirUtils.ramachandran_plot_sampled(sum(phi_mean_list,[]),sum(psi_mean_list,[]),"{}/Angles_MEANS_Predictions_NoGAPS".format(results_dir),"Sampled mean angles (φ,ψ) (without GAPS)")
    # DraupnirUtils.ramachandran_plot_sampled(sum(phi_kappa_list,[]),sum(psi_kappa_list,[]),"{}/Angles_KAPPAS_Predictions_NoGAPS".format(results_dir),"Sampled kappas angles (φ,ψ) (without GAPS)",plot_kappas=True)
def save_ancestors_predictions(name,dataset_test,aa_sequences_predictions,n_samples,results_directory,correspondence_dict,aa_prob):
    """Saves the ancestors sequences to a fasta format
    :param str name: data set project name
    :param tensor dataset_test: tensor containing the true aminoacid sequences in integer form [n_test,align_length] (gap or aa are referred from numbers 0 to 20)
    :param tensor aa_sequences_predictions: tensor containing the predicted aminoacid sequences in integer form [n_samples,n_test,align_length] (gap or aa are referred from numbers 0 to 20)
    :param int n_samples: number of samples
    :param str results_directory: path to results directory
    :param dict correspondence_dict: dictionary that contains the correspondence upon the node indexes (integers) and the tree nodes names (ordered in tree level order)
    :param aa_prob: amino acid probabilities
    """
    print("Saving ancestor's nodes sequences ...")
    node_info = dataset_test[:, 0, 1].repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(dataset_test), 1)
    node_names = ["{}//{}".format(correspondence_dict[index], index) for index in dataset_test[:, 0, 1].tolist()]
    aa_sequences_predictions = torch.cat(( node_info, aa_sequences_predictions), dim=2)
    aa_dict = DraupnirUtils.aminoacid_names_dict(aa_prob)
    aa_dict = {float(v): k for k, v in aa_dict.items()}
    aa_sequences = aa_sequences_predictions[:, :, 1:].permute(1, 0,2)  # [n_nodes,n_samples,L] #TODO: Review this is correct, the order, seems correct
    aa_sequences = np.vectorize(aa_dict.get)(aa_sequences.cpu().numpy())
    file_name = "{}/{}_sampled_nodes_seq.fasta".format(results_directory, name)
    with open(file_name, "a+") as f:
        for node_samples, node_name in zip(aa_sequences, node_names):
            for idx, sample in enumerate(node_samples):
                f.write(">Node_{}_sample_{}\n".format(node_name, idx))
                full_seq = "".join(sample)
                n = 50
                splitted_seq = [full_seq[i:i + n] for i in range(0, len(full_seq), n)]
                for segment in splitted_seq:
                    f.write("{}\n".format(segment))
def save_ancestors_predictions_coral(name,test_ordered_nodes,aa_sequences_predictions,n_samples,results_directory,correspondence_dict,aa_prob):
    """Saves the ancestors sequences to a fasta format
     :param str name: data set project name
     :param tensor dataset_test: tensor containing the true aminoacid sequences in integer form [n_test,align_length] (gap or aa are referred from numbers 0 to 20)
     :param tensor aa_sequences_predictions: tensor containing the predicted aminoacid sequences in integer form [n_samples,n_test,align_length] (gap or aa are referred from numbers 0 to 20)
     :param int n_samples: number of samples
     :param str results_directory: path to results directory
     :param dict correspondence_dict: dictionary that contains the correspondence upon the node indexes (integers) and the tree nodes names (ordered in tree level order)
     :param aa_prob: amino acid probabilities
     """
    print("Saving ancestor's nodes sequences ...")
    node_info = test_ordered_nodes.repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(test_ordered_nodes), 1)
    node_names = ["{}//{}".format(correspondence_dict[index], index) for index in test_ordered_nodes.tolist()]
    aa_sequences_predictions = torch.cat(( node_info, aa_sequences_predictions), dim=2)
    aa_dict = DraupnirUtils.aminoacid_names_dict(aa_prob)
    aa_dict = {float(v): k for k, v in aa_dict.items()}
    aa_sequences = aa_sequences_predictions[:, :, 1:].permute(1, 0,2)  # [n_nodes,n_samples,L] #TODO: Review this is correct, the order, seems correct
    aa_sequences = np.vectorize(aa_dict.get)(aa_sequences.cpu().numpy())
    file_name = "{}/{}_sampled_nodes_seq.fasta".format(results_directory, name)
    with open(file_name, "a+") as f:
        for node_samples, node_name in zip(aa_sequences, node_names):
            for idx, sample in enumerate(node_samples):
                f.write(">Node_{}_sample_{}\n".format(node_name, idx))
                full_seq = "".join(sample)
                n = 50
                splitted_seq = [full_seq[i:i + n] for i in range(0, len(full_seq), n)]
                for segment in splitted_seq:
                    f.write("{}\n".format(segment))
def percent_id_sampled_observed(dataset_test,aa_sequences_predictions,node_info,n_samples,node_names,results_directory):
    """Fast version to calculate %ID among predictions and observed data. It also calculates the number of incorrectly predicted amino acids
    :param tensor dataset_test
    :param tensor aa_sequences_predictions
    :param tensor node_info
    :param int n_samples
    :param node_names
    :param str results_directory
    """
    align_lenght = dataset_test[:,2:,0].shape[1]
    #node_names = ["{}//{}".format(correspondence_dict[index], index) for index in Dataset_test[:, 0, 1].tolist()]
    samples_names =  ["sample_{}".format(index) for index in range(n_samples)]
    equal_aminoacids = (aa_sequences_predictions[:,:,3:]== dataset_test[:, 2:, 0]).float() #is correct #[n_samples,n_nodes,L]
    #Highlight: Incorrectly predicted sites
    incorrectly_predicted_sites = (~equal_aminoacids.bool()).float().sum(-1)
    incorrectly_predicted_sites_per_sample = np.concatenate([node_info.cpu().detach().numpy(),incorrectly_predicted_sites.cpu().detach().numpy()[:,:,np.newaxis]],axis=-1)
    np.save("{}/Incorrectly_Predicted_Sites_Fast".format(results_directory), incorrectly_predicted_sites_per_sample)
    incorrectly_predicted_sites_df = pd.DataFrame(incorrectly_predicted_sites.T.cpu().detach().numpy(),index=node_names)
    incorrectly_predicted_sites_df.columns = samples_names
    incorrectly_predicted_sites_df["Average"] =incorrectly_predicted_sites_df.mean(1).values.tolist()
    incorrectly_predicted_sites_df["Std"] = incorrectly_predicted_sites_df.std(1).values.tolist()
    incorrectly_predicted_sites_df.to_csv("{}/Incorrectly_predicted_sites_df.csv".format(results_directory), sep="\t")
    #Highlight: PERCENT ID
    equal_aminoacids = equal_aminoacids.sum(-1)/align_lenght#equal_aminoacids.sum(-1)
    percent_id_df = pd.DataFrame(equal_aminoacids.T.cpu().detach().numpy()*100, index=node_names ) #[n_nodes, n_samples]
    percent_id_df.columns = samples_names
    percent_id_df["Average"] = percent_id_df.mean(1).values.tolist()
    percent_id_df["Std"] = percent_id_df.std(1).values.tolist()
    percent_id_df.to_csv("{}/PercentID_df.csv".format(results_directory),sep="\t")
    return percent_id_df, incorrectly_predicted_sites_df, align_lenght
def save_predictions_to_fasta(aa_sequences_predictions,aa_prob,name,node_names,results_directory):
    """Transforms via vectorization the integer encoded aminoacids into the amino acid letters. Saves the results into a fasta file
    :param aa_sequences_predictions: tensor dataframe with the predictions
    :param int aa_prob
    :param name: data set project name
    :param node_names: list of the real nodes names in the same order as in the dataset
    :param str results_directory"""
    aa_dict = DraupnirUtils.aminoacid_names_dict(aa_prob)
    aa_dict  = {float(v): k for k, v in aa_dict.items()}

    aa_sequences = aa_sequences_predictions[:,:,3:].permute(1,0,2) #[n_nodes,n_samples,L] #TODO: Review this is correct, the order, seems correct
    aa_sequences = np.vectorize(aa_dict.get)(aa_sequences.cpu().numpy())
    n_samples = aa_sequences.shape[1]
    file_name = "{}/{}_sampled_nodes_seq.fasta".format(results_directory,name)
    with open(file_name, "a+") as f:
        for node_samples,node_name in zip(aa_sequences,node_names):
            for idx, sample in enumerate(node_samples):
                f.write(">Node_{}_sample_{}\n".format(node_name, idx))
                full_seq = "".join(sample)
                n=50
                #splitted_seq = list(map(''.join, zip(*[iter(full_seq)]*50))) #if the seq is smaller it leaves it out
                splitted_seq = [full_seq[i:i+n] for i in range(0, len(full_seq), n)]
                for segment in splitted_seq:
                    f.write("{}\n".format(segment))
def plot_heatmap_and_incorrect_aminoacids(name,dataset_test,aa_sequences_predictions,n_samples,results_directory,correspondence_dict,aa_probs,additional_load,additional_info,replacement_plots=True):
    """Prepares the output from Draupnir for plotting by re-adding the nodes names to the data sets. It also detects large data sets and selects only
     a few nodes to plot
     :param str name: data set project name
     :param tensor dataset_test: true data [n_nodes,align_len + 2, 30]
     :param tensor aa_sequences_predictions: predictions [n_samples,n_nodes,align_len]
     :param int n_samples
     :param str results_directory
     :param dict correspondence_dict: contains the information about which node index corresponds to which node name
     :param int aa_probs
     :param namedtuple additional_load
     :param namedtuple additional_info
     :param bool replacement_plots: True --> plots the amino acid replacements occurred per site and per sequence"""
    #TODO: By simple permutation it rearranges to [n_seq,n_samples], preserving the right order?
    len_info = dataset_test[:, 0, 0].repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(dataset_test), 1)
    node_info = dataset_test[:, 0, 1].repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(dataset_test), 1)
    distance_info = dataset_test[:, 0, 2].repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(dataset_test), 1)
    node_names = ["{}//{}".format(correspondence_dict[index], index) for index in dataset_test[:, 0, 1].tolist()]
    aa_sequences_predictions = torch.cat((len_info, node_info, distance_info, aa_sequences_predictions), dim=2)
    # def Percent_ID_SAMPLED_OBSERVED():
    #     "Fast version to calculate %ID among predictions and observed data"
    #     align_lenght = dataset_test[:,2:,0].shape[1]
    #     #node_names = ["{}//{}".format(correspondence_dict[index], index) for index in Dataset_test[:, 0, 1].tolist()]
    #     samples_names =  ["sample_{}".format(index) for index in range(n_samples)]
    #     equal_aminoacids = (aa_sequences_predictions[:,:,3:]== dataset_test[:, 2:, 0]).float() #is correct #[n_samples,n_nodes,L]
    #     #Highlight: Incorrectly predicted sites
    #     incorrectly_predicted_sites = (~equal_aminoacids.bool()).float().sum(-1)
    #     incorrectly_predicted_sites_per_sample = np.concatenate([node_info.cpu().detach().numpy(),incorrectly_predicted_sites.cpu().detach().numpy()[:,:,np.newaxis]],axis=-1)
    #     np.save("{}/Incorrectly_Predicted_Sites_Fast".format(results_directory), incorrectly_predicted_sites_per_sample)
    #     incorrectly_predicted_sites_df = pd.DataFrame(incorrectly_predicted_sites.T.cpu().detach().numpy(),index=node_names)
    #     incorrectly_predicted_sites_df.columns = samples_names
    #     incorrectly_predicted_sites_df["Average"] =incorrectly_predicted_sites_df.mean(1).values.tolist()
    #     incorrectly_predicted_sites_df["Std"] = incorrectly_predicted_sites_df.std(1).values.tolist()
    #     incorrectly_predicted_sites_df.to_csv("{}/Incorrectly_predicted_sites_df.csv".format(results_directory), sep="\t")
    #     #Highlight: PERCENT ID
    #     equal_aminoacids = equal_aminoacids.sum(-1)/align_lenght#equal_aminoacids.sum(-1)
    #     percent_id_df = pd.DataFrame(equal_aminoacids.T.cpu().detach().numpy()*100, index=node_names ) #[n_nodes, n_samples]
    #     percent_id_df.columns = samples_names
    #     percent_id_df["Average"] = percent_id_df.mean(1).values.tolist()
    #     percent_id_df["Std"] = percent_id_df.std(1).values.tolist()
    #     percent_id_df.to_csv("{}/PercentID_df.csv".format(results_directory),sep="\t")
    #     return percent_id_df, incorrectly_predicted_sites_df, align_lenght

    percent_id_df,incorrectly_predicted_sites_df, alignment_length =percent_id_sampled_observed(dataset_test,aa_sequences_predictions,node_info,n_samples,node_names,results_directory)


    # def save_predictions_to_fasta(aa_sequences_predictions):
    #     aa_dict = DraupnirUtils.aminoacid_names_dict(aa_probs)
    #     aa_dict  = {float(v): k for k, v in aa_dict.items()}
    #
    #     aa_sequences = aa_sequences_predictions[:,:,3:].permute(1,0,2) #[n_nodes,n_samples,L] #TODO: Review this is correct, the order, seems correct
    #     aa_sequences = np.vectorize(aa_dict.get)(aa_sequences.cpu().numpy())
    #     n_samples = aa_sequences.shape[1]
    #     file_name = "{}/{}_sampled_nodes_seq.fasta".format(results_directory,name)
    #     with open(file_name, "a+") as f:
    #         for node_samples,node_name in zip(aa_sequences,node_names):
    #             for idx, sample in enumerate(node_samples):
    #                 f.write(">Node_{}_sample_{}\n".format(node_name, idx))
    #                 full_seq = "".join(sample)
    #                 n=50
    #                 #splitted_seq = list(map(''.join, zip(*[iter(full_seq)]*50))) #if the seq is smaller it leaves it out
    #                 splitted_seq = [full_seq[i:i+n] for i in range(0, len(full_seq), n)]
    #                 for segment in splitted_seq:
    #                     f.write("{}\n".format(segment))

    save_predictions_to_fasta(aa_sequences_predictions, aa_probs, name, node_names, results_directory)

    aa_sequences_predictions = aa_sequences_predictions.cpu().numpy()
    dataset_test = dataset_test.cpu()
    n_test = dataset_test.shape[0]
    if n_test > 100:
        torch.manual_seed(0)
        percentage_test = 50/n_test #get 50 seq only
        test_indx = (torch.rand(size=(n_test,)) < percentage_test).int().bool() #create n  (n= n internal nodes) random True and False, Supposedly 45% of the values will be 1(True)
        dataset_test = dataset_test[test_indx]
        nodes_indexes = dataset_test[:, 0, 1]
        dataset_children_predicted = aa_sequences_predictions[:,np.isin(aa_sequences_predictions[0, :, 1], nodes_indexes)]
        children_indexes_str = nodes_indexes.numpy().astype(str).tolist()
        incorrectly_predicted_sites_df = incorrectly_predicted_sites_df[incorrectly_predicted_sites_df.index.str.split("//").str[1].isin(children_indexes_str)]


    else:
        nodes_indexes = dataset_test[:, 0, 1]  #test indexes
        dataset_children_predicted = aa_sequences_predictions[:,np.isin(aa_sequences_predictions[0, :, 1], nodes_indexes)]

    plot_incorrectly_predicted_aa(incorrectly_predicted_sites_df,results_directory,alignment_length,additional_load)

    DraupnirUtils.heatmaps(dataset_children_predicted,
                           dataset_test,
                           name,
                           n_samples,
                           nodes_indexes,
                           results_directory,
                           aa_probs,
                           additional_load,
                           additional_info,
                           correspondence_dict)
    if replacement_plots:
        barplot_aa_replacement(dataset_children_predicted[:,:,3:],
                                            dataset_test[:, 2:, 0],
                                            additional_load.full_name,
                                            n_samples,
                                            nodes_indexes,
                                            results_directory,
                                            aa_probs,
                                            correspondence_dict)
def plot_entropies(name,seq_entropies,results_directory,correspondence_dict):
    """Plot amino acid entropies per alignment site
    :param str name
    :param tensor seq_entropies
    :param str results_directory
    :param dict correspondence_dict"""
    max_entropy_per_site = -((1 / 21) * np.log(1/21)) * 21  # maximum overall possible entropy per site (not sequence!)

    min_value = np.min(seq_entropies[:,1:])
    max_value = np.max(seq_entropies[:,1:])
    fig = plt.figure(figsize=(14, 8))
    ax = fig.add_subplot(111)
    folder_name = os.path.basename(results_directory)
    data_name = ["Internal nodes/Test" if folder_name.startswith("Test")  else "Leaves nodes/Train"][0]
    node_indexes = seq_entropies[:,0].tolist()
    colors = pl.cm.jet(np.linspace(0, 1, len(node_indexes)))
    ax.hlines(max_entropy_per_site, linestyles="dashed", xmin=0, xmax=seq_entropies.shape[1],color="black",label="max allowed \n entropy")
    for idx,node in enumerate(node_indexes):
        pl.plot(seq_entropies[idx,1:], color=colors[idx],label="Node {}".format(node_indexes[idx]))
    plt.legend(loc = "upper right",bbox_to_anchor=(1.14, 0.96))
    plt.ylim((0,max_entropy_per_site + 0.3))
    plt.ylabel("Shannon entropy per site (Σ(-Pi.log(Pi))")
    plt.xlabel("Sequence Sites")
    plt.title("Shannon entropy: {}".format(data_name))
    plt.savefig("{}/Entropy.png".format(results_directory))
    plt.close()
def plot_overlapping_histogram(name,dataset_train,dataset_test, aa_sequences_predictions,n_samples,results_directory,correspondence_dict,aa_prob):
    """Plot histogram whose bars represent the percentage identity between the true ancestors vs true leaves and the predicted ancestors vs true leaves
    :param str name: dataset name project
    :param tensor dataset_train
    :param tensor dataset_test
    :param tensor aa_sequences_predictions
    :param int n_samples
    :param str results_directory
    :param dict correspondence_dict: correspondence between the tree level order indexes and the --> only useful for benchmark_randall_original_naming
    :param int aa_prob
    """
    #TODO: Fix the device problem if they should be used to see the distances between internal and leaves nodes
    children_indexes = dataset_test[:, 0, 1].cpu()
    len_info = dataset_test[:, 0, 0].repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(dataset_test), 1)
    node_info = dataset_test[:, 0, 1].repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(dataset_test), 1)
    distance_info = dataset_test[:, 0, 2].repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(dataset_test), 1)
    aa_sequences_predictions = aa_sequences_predictions.cpu().detach()
    aa_sequences_predictions = torch.cat((len_info, node_info, distance_info, aa_sequences_predictions), dim=2)
    aa_sequences_predictions = aa_sequences_predictions.cpu().detach().numpy()
    dataset_children_predicted = aa_sequences_predictions[:,np.isin(aa_sequences_predictions[0, :, 1], children_indexes)]
    DraupnirUtils.build_dataframes_overlapping_histograms(dataset_children_predicted,
                                                         dataset_train.cpu(),
                                                         dataset_test.cpu(),
                                                         name,
                                                         n_samples,
                                                         children_indexes,
                                                         results_directory,
                                                         aa_prob,
                                                         correspondence_dict)
def plot_latent_space_tsne_by_clade(latent_space, additional_load, epoch, results_dir, triTSNE):
    """t-SNE projection of a z-dimensional latent space onto a 2D space. The latent space represents the sequences in the tree. The latent space is coloured according to
    the clade membership
    https://towardsdatascience.com/visualizing-feature-vectors-embeddings-using-pca-and-t-sne-ef157cea3a42
    :param tensor latent_space: [n_leaves + n_internal,1+ z_dim], first column contains the nodes indexes
    :param namedtuple additional_load
    :param int epoch
    :param str results_dir
    :param bool triTSNE: boolean paramter to indicate whether to perform a 3D latent space projection, Not used right now
    """
    # Create a two dimensional t-SNE projection of the z dim latent space
    print("Building T-SNE plot by clades")
    #annotate = [True if latent_space.shape[0] < 100 else False][0]
    annotate = False
    stripped = True
    if stripped: #no legend, no axis, nothing
        clades_dict_all = additional_load.clades_dict_all
        #n_cols = DraupnirUtils.Define_batch_size(latent_space.shape[0], batch_size=False,benchmarking=True)

        tsne_proj = TSNE(n_components=2).fit_transform(latent_space[:, 1:])
        color_map21 = matplotlib.colors.ListedColormap(
            ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
             "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
             "black"])

        color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0] #nipy_spectral, gist_rainbow
        clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
        #n_cols = int(len(clades_dict_all.keys()) / 30) + 1

        fig, ax = plt.subplots(figsize=(22, 15),dpi=200)

        for idx,(clade, nodes) in enumerate(clades_dict_all.items()):
                sequences = nodes["internal"] + nodes["leaves"]
                indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
                ax.scatter(tsne_proj[indexes, 0], tsne_proj[indexes, 1], color=clrs(idx), label=clade.replace("_"," "), alpha=1,s=700)
                if annotate:
                    for name, point in zip(sequences, tsne_proj[indexes]):
                        ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        #plt.legend(loc='upper left', prop={'size': 25},ncol=1, shadow=True)
        #plt.legend( bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 25}, ncol=1, shadow=True)
        plt.tight_layout()
        plt.axis("off")
    else:
        clades_dict_all = additional_load.clades_dict_all
        # n_cols = DraupnirUtils.Define_batch_size(latent_space.shape[0], batch_size=False,benchmarking=True)

        tsne_proj = TSNE(n_components=2).fit_transform(latent_space[:, 1:])
        color_map21 = matplotlib.colors.ListedColormap(
            ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
             "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
             "black"])

        color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0]
        clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
        # n_cols = int(len(clades_dict_all.keys()) / 30) + 1

        fig, ax = plt.subplots(figsize=(22, 15), dpi=200)

        for idx, (clade, nodes) in enumerate(clades_dict_all.items()):
            sequences = nodes["internal"] + nodes["leaves"]
            indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
            ax.scatter(tsne_proj[indexes, 0], tsne_proj[indexes, 1], color=clrs(idx), label=clade.replace("_", " "),
                       alpha=1, s=200)
            if annotate:
                for name, point in zip(sequences, tsne_proj[indexes]):
                    ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        plt.legend(title='Clades', bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 15},ncol=1,shadow=True,fontsize=15)
        plt.title("T-SNE projection of the tree's latent space; \n" + "{}".format(additional_load.full_name),fontsize=20)
        plt.axis("off")
    plt.savefig("{}/t_SNE_z_space_by_clade_epoch_{}.png".format(results_dir, epoch))
def plot_latent_space_tsne_by_clade_leaves(latent_space, additional_load, epoch, results_dir, triTSNE):
    """t-SNE projection of a z-dimensional latent space onto a 2D space.
    The latent space represents the sequences in the tree. The latent space is coloured according to
    the clade membership
    https://towardsdatascience.com/visualizing-feature-vectors-embeddings-using-pca-and-t-sne-ef157cea3a42
    :param tensor latent_space [n_leaves, 1 + z_dim]
    :param namedtuple additional_load
    :param int epoch
    :param str results_dir
    :param bool triTSNE: boolean paramter to indicate whether to perform a 3D latent space projection, Not used right now
    """
    # Create a two dimensional t-SNE projection of the z dim latent space
    print("Building t-SNE plot COLOURED by clades (only leaves)")
    #annotate = [True if latent_space.shape[0] < 100 else False][0]
    annotate = False
    stripped = True
    clades_dict_all = additional_load.clades_dict_all
    #n_cols = DraupnirUtils.Define_batch_size(latent_space.shape[0], batch_size=False,benchmarking=True)
    tsne_proj = TSNE(n_components=2).fit_transform(latent_space[:, 1:])
    color_map21 = matplotlib.colors.ListedColormap(
        ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
         "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
         "black"])

    color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0]
    clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
    #n_cols = int(len(clades_dict_all.keys()) / 30) + 1
    if stripped:
        fig, ax = plt.subplots(figsize=(22, 15))

        for idx, (clade, nodes) in enumerate(clades_dict_all.items()):
            sequences = nodes["leaves"]
            indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
            ax.scatter(tsne_proj[indexes, 0], tsne_proj[indexes, 1], color=clrs(idx), label=clade, alpha=1, s=700)
            if annotate:
                for name, point in zip(sequences, tsne_proj[indexes]):
                    ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        #plt.legend(loc='upper left', prop={'size': 25},ncol=1, shadow=True)
        plt.tight_layout()
        plt.axis("off")
    else:
        fig, ax = plt.subplots(figsize=(22, 15))

        for idx,(clade, nodes) in enumerate(clades_dict_all.items()):
                sequences = nodes["leaves"]
                indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
                ax.scatter(tsne_proj[indexes, 0], tsne_proj[indexes, 1], color=clrs(idx), label=clade, alpha=1)
                if annotate:
                    for name, point in zip(sequences, tsne_proj[indexes]):
                        ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        plt.legend(title='Clades', bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 10},ncol=1, shadow=True,fontsize=10)
        plt.title("UMAP projection of the tree's latent space; \n" + "{}".format(additional_load.full_name),fontsize=20)
    plt.savefig("{}/UMAP_z_space_by_clade__only_leaves_epoch_{}.png".format(results_dir, epoch))
    #plt.savefig("{}/Tree latent space representation (T-SNE projection); SH3 domain 200 leaves simulation".format(results_dir))
def plot_latent_space_umap_by_clade(latent_space, additional_load, epoch, results_dir, triTSNE):
    """UMAP projection of a z-dimensional latent space onto a 2D space. The latent space represents the sequences in the tree. The latent space is coloured according to
    the clade membership
    :param tensor latent_space: [n_leaves + n_internal,1+ z_dim], first column contains the nodes indexes
    :param namedtuple additional_load
    :param int epoch
    :param str results_dir
    :param bool triTSNE: boolean paramter to indicate whether to perform a 3D latent space projection, Not used right now
    """
    # Create a two dimensional t-SNE projection of the z dim latent space
    print("Building UMAP plot by clades (both internal and leaves)")
    #annotate = [True if latent_space.shape[0] < 100 else False][0]
    annotate = False
    stripped = False
    reducer = umap.UMAP()
    if stripped:
        clades_dict_all = additional_load.clades_dict_all
        #n_cols = DraupnirUtils.Define_batch_size(latent_space.shape[0], batch_size=False,benchmarking=True)
        tsne_proj = reducer.fit_transform(latent_space[:, 1:])

        color_map21 = matplotlib.colors.ListedColormap(
            ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
             "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
             "black"])

        color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0] #nipy_spectral, gist_rainbow
        clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
        #n_cols = int(len(clades_dict_all.keys()) / 30) + 1

        fig, ax = plt.subplots(figsize=(22, 15),dpi=200)

        for idx,(clade, nodes) in enumerate(clades_dict_all.items()):
                sequences = nodes["internal"] + nodes["leaves"]
                indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
                ax.scatter(tsne_proj[indexes, 0], tsne_proj[indexes, 1], color=clrs(idx), label=clade.replace("_"," "), alpha=1,s=700)
                if annotate:
                    for name, point in zip(sequences, tsne_proj[indexes]):
                        ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        #plt.legend(loc='upper left', prop={'size': 25},ncol=1, shadow=True)
        #plt.legend( bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 25}, ncol=1, shadow=True)
        plt.tight_layout()
        plt.axis("off")
    else:
        clades_dict_all = additional_load.clades_dict_all
        # n_cols = DraupnirUtils.Define_batch_size(latent_space.shape[0], batch_size=False,benchmarking=True)

        tsne_proj = reducer.fit_transform(latent_space[:, 1:])
        color_map21 = matplotlib.colors.ListedColormap(
            ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
             "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
             "black"])

        color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0]
        clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
        # n_cols = int(len(clades_dict_all.keys()) / 30) + 1

        fig, ax = plt.subplots(figsize=(22, 15), dpi=200)

        for idx, (clade, nodes) in enumerate(clades_dict_all.items()):
            sequences = nodes["internal"] + nodes["leaves"]
            indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
            ax.scatter(tsne_proj[indexes, 0], tsne_proj[indexes, 1], color=clrs(idx), label=clade.replace("_", " "),
                       alpha=1, s=200)
            if annotate:
                for name, point in zip(sequences, tsne_proj[indexes]):
                    ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        plt.legend(title='Clades', bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 15},ncol=1,shadow=True,fontsize=15)
        plt.title("UMAP projection of the tree's latent space; \n" + "{}".format(additional_load.full_name),fontsize=20)
        plt.axis("off")
    plt.savefig("{}/UMAP_z_space_by_clade_epoch_{}.png".format(results_dir, epoch))
def plot_latent_space_umap_by_clade_leaves(latent_space, additional_load, epoch, results_dir, triTSNE):
    """UMAP projection of a z-dimensional latent space onto a 2D space. The latent space represents the sequences in the tree. The latent space is coloured according to
    the clade membership
    :param tensor latent_space: [n_leaves,1+ z_dim], first column contains the nodes indexes
    :param namedtuple additional_load
    :param int epoch
    :param str results_dir
    :param bool triTSNE: boolean paramter to indicate whether to perform a 3D latent space projection, Not used right now
    """
    # Create a two dimensional t-SNE projection of the z dim latent space
    print("Building UMAP plot COLOURED by clades")
    #annotate = [True if latent_space.shape[0] < 100 else False][0]
    annotate = False
    stripped = True
    clades_dict_all = additional_load.clades_dict_all
    #n_cols = DraupnirUtils.Define_batch_size(latent_space.shape[0], batch_size=False,benchmarking=True)
    reducer = umap.UMAP()
    umap_proj =  reducer.fit_transform(latent_space[:,1:])
    #tsne_proj = TSNE(n_components=2).fit_transform(latent_space[:, 1:])
    color_map21 = matplotlib.colors.ListedColormap(
        ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
         "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
         "black"])

    color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0]
    clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
    #n_cols = int(len(clades_dict_all.keys()) / 30) + 1
    if stripped:
        fig, ax = plt.subplots(figsize=(22, 15))

        for idx, (clade, nodes) in enumerate(clades_dict_all.items()):
            sequences = nodes["leaves"]
            indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
            ax.scatter(umap_proj[indexes, 0], umap_proj[indexes, 1], color=clrs(idx), label=clade, alpha=1, s=700)
            if annotate:
                for name, point in zip(sequences, umap_proj[indexes]):
                    ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        #plt.legend(loc='upper left', prop={'size': 25},ncol=1, shadow=True)
        plt.tight_layout()
        plt.axis("off")
    else:
        fig, ax = plt.subplots(figsize=(22, 15))

        for idx,(clade, nodes) in enumerate(clades_dict_all.items()):
                sequences = nodes["leaves"]
                indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
                ax.scatter(umap_proj[indexes, 0], umap_proj[indexes, 1], color=clrs(idx), label=clade, alpha=1)
                if annotate:
                    for name, point in zip(sequences, umap_proj[indexes]):
                        ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        plt.legend(title='Clades', bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 10},ncol=1, shadow=True,fontsize=10)
        plt.title("UMAP projection of the tree's latent space; \n" + "{}".format(additional_load.full_name),fontsize=20)
    plt.savefig("{}/UMAP_z_space_by_clade__only_leaves_epoch_{}.png".format(results_dir, epoch))
    #plt.savefig("{}/Tree latent space representation (T-SNE projection); SH3 domain 200 leaves simulation".format(results_dir))
def plot_latent_space_pca_by_clade(latent_space,additional_load,num_epochs, results_dir):
    """PCA projection of a z-dimensional latent space onto a 2D space. The latent space represents the sequences in the tree. The latent space is coloured according to
    the clade membership
    :param tensor latent_space: [n_leaves + n_internal,1+ z_dim], first column contains the nodes indexes
    :param namedtuple additional_load
    :param int epoch
    :param str results_dir
    :param bool triTSNE: boolean paramter to indicate whether to perform a 3D latent space projection, Not used right now
    """
    print("Building PCA plot")
    #annotate = [True if latent_space.shape[0] < 100 else False][0]
    annotate = False
    clades_dict_all = additional_load.clades_dict_all
    pca = PCA(n_components=2)
    pca.fit(latent_space[:,1:])
    latent_space_pca = pca.transform(latent_space[:,1:])
    color_map21 = matplotlib.colors.ListedColormap(
        ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
         "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
         "black"])

    color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0]
    clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
    #n_cols = int(len(clades_dict_all.keys()) / 30) + 1
    stripped = True
    if stripped:
        fig, ax = plt.subplots(figsize=(22, 15))

        for idx, (clade, nodes) in enumerate(clades_dict_all.items()):
            sequences = nodes["internal"] + nodes["leaves"]
            indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
            ax.scatter(latent_space_pca[indexes, 0], latent_space_pca[indexes, 1], color=clrs(idx), label=clade, alpha=1, s=200)
            if annotate:
                for name, point in zip(sequences, latent_space_pca[indexes]):
                    ax.annotate(name, xy=(point[0], point[1]), size=7)  # xytext=(1,1)
        #plt.legend(loc='upper left', prop={'size': 25},ncol=1, shadow=True)
        plt.legend(bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 25}, ncol=1, shadow=True)

        plt.tight_layout()
        plt.axis("off")
    else:
        fig, ax = plt.subplots(figsize=(22, 15))
        for idx, (clade, nodes) in enumerate(clades_dict_all.items()):
            sequences = nodes["internal"] + nodes["leaves"]
            indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
            ax.scatter(latent_space_pca[indexes, 0], latent_space_pca[indexes, 1], color=clrs(idx), label=clade, alpha=0.5)
            if annotate:
                for name, point in zip(sequences, latent_space_pca[indexes]):
                    ax.annotate(name, xy=(point[0], point[1]))  # xytext=(1,1)
        plt.xlabel("PC 1", fontsize = 20)
        plt.ylabel("PC 2", fontsize = 20)
        plt.legend(title='Clades', bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 10},ncol=1,shadow=True,fontsize=10)
        plt.title("PCA projection of the tree's latent space;\n" + r"{}".format(additional_load.full_name),fontsize=20)
    plt.savefig("{}/PCA_z_space_epoch_{}.png".format(results_dir, num_epochs))
def plot_latent_space_pca_by_clade_leaves(latent_space,additional_load,num_epochs, results_dir):
    """PCA projection of a z-dimensional latent space onto a 2D space. The latent space represents the sequences in the tree. The latent space is coloured according to
    the clade membership
    https://towardsdatascience.com/visualizing-feature-vectors-embeddings-using-pca-and-t-sne-ef157cea3a42
    :param tensor latent_space: [n_leaves,1+ z_dim], first column contains the nodes indexes
    :param namedtuple additional_load
    :param int epoch
    :param str results_dir
    :param bool triTSNE: boolean paramter to indicate whether to perform a 3D latent space projection, Not used right now
    """
    """"""
    print("Building PCA plot")
    #annotate = [True if latent_space.shape[0] < 100 else False][0]
    annotate = False
    clades_dict_all = additional_load.clades_dict_all
    pca = PCA(n_components=2)
    pca.fit(latent_space[:,1:])
    latent_space_pca = pca.transform(latent_space[:,1:])
    color_map21 = matplotlib.colors.ListedColormap(
        ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
         "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
         "black"])

    color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0]
    clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
    #n_cols = int(len(clades_dict_all.keys()) / 30) + 1


    fig, ax = plt.subplots(figsize=(22, 15))

    for idx, (clade, nodes) in enumerate(clades_dict_all.items()):
        sequences =  nodes["leaves"]
        indexes = (latent_space[:, 0][..., None] == sequences).any(-1)
        ax.scatter(latent_space_pca[indexes, 0], latent_space_pca[indexes, 1], color=clrs(idx), label=clade, alpha=0.5)
        if annotate:
            for name, point in zip(sequences, latent_space_pca[indexes]):
                ax.annotate(name, xy=(point[0], point[1]))  # xytext=(1,1)
    plt.xlabel("PC 1", fontsize = 20)
    plt.ylabel("PC 2", fontsize = 20)
    plt.legend(title='Clades', bbox_to_anchor=(1.01, 1), loc='upper left', prop={'size': 10},ncol=1,shadow=True,fontsize=10)
    plt.title("PCA projection of the tree's latent space;\n" + r"{}".format(additional_load.full_name),fontsize=20)
    plt.savefig("{}/PCA_z_space_epoch_{}.png".format(results_dir, num_epochs))
def plot_pairwise_distances(latent_space,additional_load,num_epochs, results_dir):
    """Plots the pairwise distance between the latent space vectors of 2 nodes in the tree vs the branch length/patristic distance between them
    :param tensor latent_space
    :param namedtuple additional_load
    :param int num_epochs
    :param str results_dir"""
    print("Plotting z distances vs branch lengths...")
    clades_dict_all = additional_load.clades_dict_all

    color_map21 = matplotlib.colors.ListedColormap(
        ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
         "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
         "black"])
    color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0]  # nipy_spectral, gist_rainbow
    clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
    linked_nodes_dict = additional_load.linked_nodes_dict
    patristic_matrix_full = additional_load.patristic_matrix_full
    patristic_matrix_full_no_indexes = patristic_matrix_full[1:,1:]
    latent_space_no_indexes = latent_space[:,1:]
    use_cosine_similarity = False
    def cosine_similarity(a,b):
        return  np.dot(a, b)/(np.linalg.norm(a)*np.linalg.norm(b))
    def pairwise_dist(a,b):
        return np.linalg.norm(a-b)
    if use_cosine_similarity:
        distance_type = "Cosine_similarity"
        distance_fn = cosine_similarity
    else:
        distance_type = "Pairwise_distance"
        distance_fn = pairwise_dist
    fig, ax = plt.subplots(figsize=(22, 15), dpi=200)
    distances_list = []
    for node, children in linked_nodes_dict.items():
        if children: #if it has descendants
            # Highlight: find to which clade does this node belong and use its color---> List comprehesion does not work
            for idx, (clades, nodes) in enumerate(clades_dict_all.items()):
                if node in nodes["internal"] or node in nodes["leaves"]:
                    color_clade = clrs(idx)
            for child in children:
                nodes_pairs = torch.tensor([float(node),float(child)])
                pair_indexes_branch_lengths = (patristic_matrix_full[1:, 0][..., None] == nodes_pairs.cpu()).any(-1)
                pair_branch_lengths = patristic_matrix_full_no_indexes[pair_indexes_branch_lengths]
                pair_branch_lengths = pair_branch_lengths[:,pair_indexes_branch_lengths][0,1].detach().cpu().numpy()
                pair_indexes_latent_space = (latent_space[:,0][..., None] == nodes_pairs).any(-1)
                vector_pair_latent_space = latent_space_no_indexes[pair_indexes_latent_space].detach().cpu().numpy()
                #latent_space_pairwise_distance = np.linalg.norm(vector_pair_latent_space[0] - vector_pair_latent_space[1])
                latent_space_distance = distance_fn(vector_pair_latent_space[0], vector_pair_latent_space[1])

                distances_list.append(np.array([pair_branch_lengths,latent_space_distance]).T)
                ax.scatter(pair_branch_lengths,latent_space_distance,color = color_clade, s=200)
    distances_array = np.vstack(distances_list)
    correlation_coefficient = np.corrcoef(distances_array,rowvar=False)
    #plt.ylabel("Z vector distance between contiguous nodes",fontsize=20)
    plt.ylabel("Euclidean distance", fontsize=20)
    plt.xlabel("Branch length between contiguous nodes",fontsize=20)
    plt.title("GP-VAE: Z {} vs Branch lengths for linked nodes (ancestors and leaves) \n Correlation Coefficient: {}".format(distance_type,correlation_coefficient[0,1]),fontsize=20)
    plt.savefig("{}/Distances_GP_VAE_z_vs_branch_lengths_{}_INTERNAL_and_LEAVES.png".format(results_dir,distance_type))

    # distances_array = np.vstack(distances_list)
    # correlation_coefficient = np.corrcoef(distances_array,rowvar=False)
    # with open("{}/correlation_coeff.txt".format(results_dir),"a") as the_file:
    #     the_file.write('Correlation coefficient between branch lengths and latent space: {}\n'.format(correlation_coefficient[0,1]))

def plot_pairwise_distances_only_leaves(latent_space,additional_load,num_epochs, results_dir,patristic_matrix_train):
    """Plots the distance between the latent space vectors of 2 leaf nodes in the tree and the branch length between them
    :param tensor latent_space [n_leaves, 1 + z_dim]
    :param namedtuple additional_load
    :param int num_epochs
    :param str results_dir
    :param patristic_matrix_train : [n_leaves + 1, n_leaves + 1]"""
    print("Plotting z distances vs branch lengths...")
    clades_dict_all = additional_load.clades_dict_all

    color_map21 = matplotlib.colors.ListedColormap(
        ["plum", "navy", "turquoise", "peachpuff", "palevioletred", "red", "darkorange", "yellow", "lime", "green",
         "dodgerblue", "blue", "purple", "magenta", "grey", "maroon", "lightcoral", "olive", "teal", "goldenrod",
         "black"])
    color_map_name = [color_map21 if len(clades_dict_all) <= 21 else 'nipy_spectral'][0]  # nipy_spectral, gist_rainbow
    clrs = plt.get_cmap(color_map_name, len(clades_dict_all))
    patristic_matrix_train = patristic_matrix_train.detach().cpu()
    patristic_matrix_train_no_indexes = patristic_matrix_train[1:,1:]
    patristic_matrix_train_no_top_indexes = patristic_matrix_train[1:,:].detach().cpu().numpy()
    latent_space_no_indexes = latent_space[:,1:].detach().cpu().numpy()
    latent_space_indexes = latent_space[:,0]
    latent_space = latent_space.detach().cpu()
    #proj = TSNE(n_components=2).fit_transform(latent_space[:, 1:])
    #reducer = umap.UMAP()
    #proj = reducer.fit_transform(latent_space[:, 1:])
    fig, ax = plt.subplots(figsize=(22, 15), dpi=200)
    distances_list = []
    use_cosine_similarity = False
    def cosine_similarity(a,b):
        return  np.dot(a, b)/(np.linalg.norm(a)*np.linalg.norm(b))
    def pairwise_dist(a,b):
        return np.linalg.norm(a-b)

    if use_cosine_similarity:
        distance_type = "Cosine_similarity"
        distance_fn = cosine_similarity
    else:
        distance_type = "Pairwise_distance"
        distance_fn = pairwise_dist

    assert latent_space[:,0].all() == patristic_matrix_train[1:,0].all()
    nodes_indexes = latent_space[:,0]
    for leaf_i_z, leaf_j_z,leaf_i_idx,leaf_j_idx in zip(latent_space_no_indexes,latent_space_no_indexes[1:],nodes_indexes,nodes_indexes[1:]):
            for idx, (clade, nodes) in enumerate(clades_dict_all.items()):
                if leaf_i_idx in nodes["leaves"]:
                    color_clade = clrs(idx)
            latent_space_distance = distance_fn(leaf_i_z, leaf_j_z)
            nodes_pairs = torch.tensor([leaf_i_idx,leaf_j_idx]).cpu()
            print("node pairs")
            print(nodes_pairs)
            print("patristic matrix train")
            print(patristic_matrix_train[1:,0])
            branch_length_idx = (patristic_matrix_train[1:,0][..., None] == nodes_pairs).any(-1)
            print("branch length idx")
            print(branch_length_idx)
            branch_length = patristic_matrix_train_no_indexes[branch_length_idx]
            branch_length = branch_length[:,branch_length_idx]
            print("branch length")
            print(branch_length)
            distances_list.append(np.array([branch_length[0,1],latent_space_distance]).T)
            ax.scatter(branch_length[0,1],latent_space_distance,color = color_clade, s=200)
    distances_array = np.vstack(distances_list)
    pearson_correlation_coefficient = np.corrcoef(distances_array,rowvar=False)
    print("Correlation coefficient: {}".format(pearson_correlation_coefficient))
    spearman_correlation_coefficient = stats.spearmanr(distances_array[:,0],distances_array[:,1])[0]
    print("Spearman correlation coefficient {}".format(spearman_correlation_coefficient))
    #plt.ylabel("Z vector pairwise distance",fontsize=50)
    plt.ylabel("Euclidean distance", fontsize=50)
    plt.xlabel("Branch length ",fontsize=50)
    plt.title("GP-VAE: Z {} vs Branch lengths for linked nodes (ancestors and leaves) \n Spearman correlation Coefficient: {}".format(distance_type,spearman_correlation_coefficient),fontsize=20)

    #plt.title("TOU-VAE: Z {} vs Branch lengths between leaves. \n Correlation coefficient : {} \n Spearman correlation : {}".format(distance_type,pearson_correlation_coefficient[0,1], spearman_correlation_coefficient),fontsize=40)
    #plt.title("Standard VAE",fontsize=50)
    #plt.title("Draupnir marginal", fontsize=50)
    plt.savefig("{}/Distances_GP_VAE_z_vs_branch_lengths_{}_ONLY_LEAVES.png".format(results_dir,distance_type))
def align_true_predicted(records,results_dir): #TODO: can be faster, although is mainly used with the coral sequences, which only have 5 sequences
    """Writes the list of records (sampled and true), transforms into fasta file
    :param list records: list of biopython Seq records with the sequences in the alignment
    :param str results_dir"""
    samples_out_file = os.path.join(results_dir,"samples_plus_true_to_align.fasta")
    SeqIO.write(records, samples_out_file, "fasta")
    alignment_out_file = os.path.join(results_dir,"samples_plus_true_aligned.fasta")
    dict_alignment, alignment = DraupnirUtils.infer_alignment(None, samples_out_file, alignment_out_file)
    alignment_length = len(alignment[0].seq)
    sampled_sequences = defaultdict()
    true_sequences = defaultdict()
    for key, value in dict_alignment.items():
        if key.startswith("Sample"):
            sampled_sequences[key] = value
        elif key.startswith("True"):
            true_sequences[key] = value
    return true_sequences,sampled_sequences,alignment_length
def clean_and_realign_test(name,dataset_test,aa_sequences_predictions,test_ordered_nodes,n_samples,aa_probs,results_dir,additional_load,additional_info):
    """
     a)Select from the predictions only the sequences in the dataset_test.
     b) Convert to sequences
     c) Remove gaps
     d) Save to file both samples and the true observed nodes (for coral, there is only 1 node)
     c) Align to the 'observed' sequences to determine percent identity
     :param str name
     :param tensor dataset_test
     :param tensor aa_sequences_predictions
     :param test_ordered_nodes
     :param int n_samples
     :param int aa_probs
     :param str results_dir
     :param namedtuple additional_load
     :param namedtuple additional_info
      """
    node_info = test_ordered_nodes.repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(test_ordered_nodes), 1)
    aa_sequences_predictions = torch.cat((node_info,aa_sequences_predictions),dim=2)
    #aa_sequences_predictions = aa_sequences_predictions.permute(1,0,2)
    test_indx_patristic = (aa_sequences_predictions[:,:, 0][..., None] == dataset_test[:,0,1]).any(-1) #[..., None]
    aa_sequences_predictions = aa_sequences_predictions[test_indx_patristic].unsqueeze(1) #unsqueeze to keep the the size = [n_samples, n_nodes, seq-len]
    n_unique_nodes = len(list(set(dataset_test[:,0,1].tolist()))) #for the coral faviina sequences we have 5 times the root node
    tree_levelorder_dict = dict(zip(list(range(0,len(additional_load.tree_levelorder_names))),additional_load.tree_levelorder_names))
    special_nodes_list = list(additional_load.special_nodes_dict.keys()) #assuming it preserves the order    exit()

    records = []
    for sample_idx in range(n_samples):
        for node_idx in range(n_unique_nodes):
            predicted_seq = aa_sequences_predictions[sample_idx,node_idx]
            predicted_seq_letters = DraupnirUtils.convert_to_letters(predicted_seq,aa_probs)
            record = SeqRecord.SeqRecord(Seq(''.join(predicted_seq_letters).replace("-", "")),
                                         annotations={"molecule_type": "protein"},
                                         id="Sampled_node_{}_{}_sample_index_{}".format(node_idx,tree_levelorder_dict[node_idx],sample_idx),
                                         description="")
            records.append(record)
    # Highlight: Append also the true internal/test sequences (from the paper)
    for true_index, true_seq in enumerate(dataset_test[:, 2:, 0]):
        true_seq_letters = DraupnirUtils.convert_to_letters(true_seq, aa_probs)
        node_number = dataset_test[true_index, 0, 1].item()  # tree levelorder number
        node_name = tree_levelorder_dict[node_number]  # tree name
        node_true_name = special_nodes_list[true_index]  # sequence name in the paper
        record = SeqRecord.SeqRecord(Seq(''.join(true_seq_letters).replace("-", "")),
                                     annotations={"molecule_type": "protein"},
                                     id="True_node_{}_{}".format(node_name, node_true_name), description="")
        records.append(record)

    # def Align():
    #     samples_out_file = os.path.join(results_dir,"samples_plus_true_to_align.fasta")
    #     SeqIO.write(records, samples_out_file, "fasta")
    #     alignment_out_file = os.path.join(results_dir,"samples_plus_true_aligned.fasta")
    #     dict_alignment, alignment = DraupnirUtils.infer_alignment(None, samples_out_file, alignment_out_file)
    #     alignment_length = len(alignment[0].seq)
    #     sampled_sequences = defaultdict()
    #     true_sequences = defaultdict()
    #     for key, value in dict_alignment.items():
    #         if key.startswith("Sample"):
    #             sampled_sequences[key] = value
    #         elif key.startswith("True"):
    #             true_sequences[key] = value
    #     return true_sequences,sampled_sequences,alignment_length
    true_sequences, sampled_sequences, alignment_length = align_true_predicted(records,results_dir)


    def plot_results():
        correlations_dataframe_TEST = pd.DataFrame(index=["Sample_{}".format(n) for n in range(n_samples)] + ["Average", "Std"])
        #HIGHLIGHT : WE HAVE several true sequences for the same node, therefore, we will check all samples for the same node to each of the true nodes
        incorrectly_predicted_sites = []
        percentage_identity_dict = defaultdict()
        blosum_score_dict = defaultdict()
        blosum_score_dict_true = defaultdict()
        for true_name, true_seq in true_sequences.items():
            blosum_true = DraupnirUtils.score_pairwise(true_seq,true_seq, additional_info.blosum_dict, gap_s=11, gap_e=1)
            blosum_score_dict_true[true_name] = blosum_true
            incorrectly_predicted_sites_sample = []
            percentage_identity_samples=[]
            blosum_score_samples = []
            for sample_name, sample_seq in sampled_sequences.items():
                ips = DraupnirUtils.incorrectly_predicted_aa(sample_seq, true_seq)
                pid = DraupnirUtils.perc_identity_pair_seq(sample_seq,true_seq) #TODO: is this correctly done?
                blosum = DraupnirUtils.score_pairwise(sample_seq,true_seq, additional_info.blosum_dict, gap_s=11, gap_e=1)
                incorrectly_predicted_sites_sample.append(ips)
                percentage_identity_samples.append(pid)
                blosum_score_samples.append(blosum)
                correlations_dataframe_TEST.loc[["Sample_{}".format(sample_name.split("_")[-1])], "PercentageID_Sampled_{}_Observed_{}".format(sample_name.split("_")[-1],true_name)] = DraupnirUtils.perc_identity_pair_seq(sample_seq,true_seq)
            incorrectly_predicted_sites.append(incorrectly_predicted_sites_sample)
            percentage_identity_dict[true_name] = sum(percentage_identity_samples)/len(percentage_identity_samples)
            blosum_score_dict[true_name] = sum(blosum_score_samples)/len(blosum_score_samples)

        incorrectly_predicted_sites_array = np.array(incorrectly_predicted_sites)
        np.save("{}/Incorrectly_Predicted_Sites".format(results_dir), incorrectly_predicted_sites)
        average_incorrectly_predicted_sites = np.mean(incorrectly_predicted_sites_array, axis=1)
        std_incorrectly_predicted_sites = np.std(incorrectly_predicted_sites_array, axis=1)
        fig = plt.figure(figsize=(14, 8))
        ax = fig.add_subplot(111)
        ## the data
        N = average_incorrectly_predicted_sites.shape[0]
        ## necessary variables
        ind = np.arange(N)  # the x locations for the groups
        width = 0.35  # the width of the bars
        blue, = sns.color_palette("muted", 1)
        ## the bars
        rects1 = ax.bar(ind, average_incorrectly_predicted_sites,
                        width,
                        color=blue,
                        alpha=0.75,
                        edgecolor="blue",
                        yerr=std_incorrectly_predicted_sites,
                        error_kw=dict(elinewidth=2, ecolor='red'))

        DraupnirUtils.autolabel(rects1, ax,blosum_dict=blosum_score_dict,percent_id_dict=percentage_identity_dict,blosum_true_dict = blosum_score_dict_true)
        # axes and labels
        ax.set_xlim(-width, N * + width)
        ax.set_ylim(0, alignment_length)
        ax.set_ylabel('Number of incorrect sites')
        ax.set_title('{} : Incorrectly predicted aa sites (%ID)'.format(name))
        xTickMarks = list(true_sequences.keys())
        ax.set_xticks(ind + width)
        xtickNames = ax.set_xticklabels(xTickMarks)
        plt.setp(xtickNames, rotation=10, fontsize=8)

        ## add a legend--> Not working (inside or outside loop)
        # ax.legend((rects1[0]), ('Draupnir')) #changed from rects1[0], rects2[0]
        plt.title(r"Incorrectly predicted amino acids; {}".format(additional_load.full_name))
        plt.savefig("{}/IncorrectlyPredictedAA_BarPlot.png".format(results_dir))
        plt.close()

        columns_test = sorted(list(correlations_dataframe_TEST.columns.values))

        correlations_dataframe_TEST = correlations_dataframe_TEST.round(1)
        correlations_dataframe_TEST = correlations_dataframe_TEST[columns_test]
        correlations_dataframe_TEST = correlations_dataframe_TEST.transpose()

        correlations_dataframe_TEST.to_csv("{}/Correlations_samples_vs_Observed_INCORRECTLY_predicted_AA.csv".format(results_dir), sep="\t")


    plot_results()
def clean_and_realign_train(name,dataset_test,dataset_train,aa_sequences_predictions,test_ordered_nodes,n_samples,aa_probs,results_dir,additional_load,additional_info):
    """Function designed to deal with data sets whose ancestral sequences aree not aligned to the leaves.
     :param str name
     :param dataset_test
     :param dataset_train
     :param aa_sequences_predictions
     :param test_ordered_nodes
     :param n_samples
     :param aa_probs
     :param results_dir
     :param additional_load
     :param additional_info"""
    # Highlight: Heatmap of the test sequences against the train sequences
    node_info = test_ordered_nodes.repeat(n_samples).unsqueeze(-1).reshape(n_samples, len(test_ordered_nodes), 1)
    aa_sequences_predictions = torch.cat((node_info,aa_sequences_predictions),dim=2)
    #aa_sequences_predictions = aa_sequences_predictions.permute(1,0,2)
    test_indx_patristic = (aa_sequences_predictions[:,:, 0][..., None] == dataset_test[:,0,1]).any(-1) #[..., None]
    aa_sequences_predictions = aa_sequences_predictions[test_indx_patristic].unsqueeze(1) #unsqueeze to keep the the size = [n_samples, n_nodes, seq-len]
    n_unique_nodes = len(list(set(dataset_test[:,0,1].tolist()))) #for the coral faviina sequences we have 5 times the root node
    tree_levelorder_dict = dict(zip(list(range(0,len(additional_load.tree_levelorder_names))),additional_load.tree_levelorder_names))
    special_nodes_list = list(additional_load.special_nodes_dict.keys()) #assuming it preserves the order    exit()

    records=[]
    #Append the test samples
    for sample_idx in range(n_samples):
        for node_idx in range(n_unique_nodes):
            predicted_seq = aa_sequences_predictions[sample_idx, node_idx]
            predicted_seq_letters = DraupnirUtils.convert_to_letters(predicted_seq, aa_probs)
            record = SeqRecord.SeqRecord(Seq(''.join(predicted_seq_letters).replace("-", "")),
                                         annotations={"molecule_type": "protein"},
                                         id="Sampled_node_{}_{}_sample_index_{}".format(node_idx,
                                                                                        tree_levelorder_dict[node_idx],
                                                                                        sample_idx),
                                         description="")
            records.append(record)

    #Append the train sequences
    for train_index, true_train_seq in enumerate(dataset_train[:, 2:, 0]):
        true_seq_letters = DraupnirUtils.convert_to_letters(true_train_seq, aa_probs)
        node_number = dataset_train[train_index, 0, 1].item()  # tree levelorder number
        node_name = tree_levelorder_dict[node_number]  # tree name
        #node_true_name = special_nodes_list[train_index]  # sequence name in the paper
        record = SeqRecord.SeqRecord(Seq(''.join(true_seq_letters).replace("-", "")),
                                     annotations={"molecule_type": "protein"},
                                     id="Train_True_node_{}_{}".format(node_name, train_index), description="")
        records.append(record)

    samples_out_file = os.path.join(results_dir, "test_samples_plus_train_true_to_align.fasta")
    SeqIO.write(records, samples_out_file, "fasta")
    alignment_out_file = os.path.join(results_dir, "test_samples_plus_train_true_aligned.fasta")
    dict_alignment, alignment = DraupnirUtils.infer_alignment(None, samples_out_file, alignment_out_file)
    alignment_length = len(alignment[0].seq)
    test_sampled_sequences = defaultdict()
    train_true_sequences = defaultdict()
    for key, value in dict_alignment.items():
        if key.startswith("Sample"):
            test_sampled_sequences[key] = value
        elif key.startswith("Train_True"):
            train_true_sequences[key] = value

    correlations_dataframe_PID = defaultdict()
    for node_idx in range(n_unique_nodes):
        nodes_seq_dict = {key: value for key, value in test_sampled_sequences.items() if "node_{}".format(node_idx) in key}
        correlations_dataframe_PID_node = pd.DataFrame(index=[name for name in nodes_seq_dict.keys()] + ["Average", "Std"])
        percentage_identity_dict = defaultdict()
        blosum_score_dict = defaultdict()
        for train_true_name, train_true_seq in train_true_sequences.items():
            percentage_identity_samples = []
            blosum_score_samples = []
            for test_sample_name, sample_seq in nodes_seq_dict.items():
                pid = DraupnirUtils.perc_identity_pair_seq(sample_seq, train_true_seq)
                blosum = DraupnirUtils.score_pairwise(sample_seq, train_true_seq, additional_info.blosum_dict, gap_s=11, gap_e=1)
                percentage_identity_samples.append(pid)
                blosum_score_samples.append(blosum)
                correlations_dataframe_PID_node.loc[[test_sample_name], "PercentageID_vs_Train_Observed_{}".format( train_true_name)] = DraupnirUtils.perc_identity_pair_seq(sample_seq,train_true_seq)

            average_pid = sum(percentage_identity_samples) / len(percentage_identity_samples)
            if n_samples > 1:
                std_pid = statistics.stdev(percentage_identity_samples)
            else:
                std_pid = [0]
            percentage_identity_dict[train_true_name] = average_pid
            blosum_score_dict[train_true_name] = sum(blosum_score_samples) / len(blosum_score_samples)
            correlations_dataframe_PID_node.loc[["Average"], "PercentageID_vs_Train_Observed_{}".format(train_true_name)] = average_pid
            correlations_dataframe_PID_node.loc[["Std"], "PercentageID_vs_Train_Observed_{}".format(train_true_name)] = std_pid
        correlations_dataframe_PID["node_{}".format(node_idx)] = correlations_dataframe_PID_node


    train_nodes_names = [name.replace("Train_True_node_","").split("_")[0] for name in train_true_sequences.keys()]

    for node_idx in range(n_unique_nodes):
            correlations_dataframe_PID_node = correlations_dataframe_PID["node_{}".format(node_idx)]
            average_pid = np.array(correlations_dataframe_PID_node.loc[["Average"]].values.tolist()).squeeze(0)
            std_pid = np.array(correlations_dataframe_PID_node.loc[["Std"]].values.tolist()).squeeze(0)
            fig = plt.figure(figsize=(14, 8))
            ax = fig.add_subplot(111)
            ## the data
            N = average_pid.shape[0]
            ## necessary variables
            ind = np.arange(N)  # the x locations for the groups
            width = 0.35  # the width of the bars
            blue, = sns.color_palette("muted", 1)
            ## the bars
            rects1 = ax.bar(ind, average_pid,
                            width,
                            color=blue,
                            alpha=0.75,
                            edgecolor="blue",
                            yerr=std_pid,
                            error_kw=dict(elinewidth=2, ecolor='red'))

            DraupnirUtils.autolabel(rects1, ax)
            # axes and labels
            ax.set_xlim(-width, N * + width)
            ax.set_ylim(0, 100)
            ax.set_ylabel('% ID ')
            ax.set_title('{} :% ID of {} against the train sequences'.format(name,node_idx))
            xTickMarks = train_nodes_names
            ax.set_xticks(ind + width)
            xtickNames = ax.set_xticklabels(xTickMarks)
            plt.setp(xtickNames, rotation=180, fontsize=8)

            ## add a legend--> Not working (inside or outside loop)
            # ax.legend((rects1[0]), ('Draupnir')) #changed from rects1[0], rects2[0]
            plt.title(r"% ID of the predicted internal node (root) against the leaves; {}".format(additional_load.full_name))
            plt.savefig("{}/PercentID_node_{}_against_train_seq.pdf".format(results_dir,node_idx))
            plt.close()
def plot_incorrectly_predicted_aa(incorrectly_predicted_sites_df,results_directory,align_len, additional_load):
    """Plots a histogram with the number of incorrectly predicted sites per sequence
    :param pandas-dataframe incorrectly_predicted_sites_df
    :param str results_directory
    :param align_len: alignment length
    :param namedtuple additional_load
    """
    print("Building Fast Incorrectly Predicted aa...")
    average_incorrectly_predicted_sites = incorrectly_predicted_sites_df["Average"].values
    try:
        accuracy = len(np.where(average_incorrectly_predicted_sites < 75)[0]) / average_incorrectly_predicted_sites.shape[0]
        print("Accuracy (% incorrect sites < 75)  : {}".format(accuracy))
    except:
        accuracy = 0.
    folder_name =os.path.basename(results_directory)
    data_name = ["Internal nodes" if folder_name.startswith("Test") else "Leaves nodes"][0]
    results_directory_abs = os.path.dirname(results_directory)

    text_file = open("{}/Hyperparameters_{}.txt".format(results_directory_abs, "_".join(results_directory_abs.split("_")[-5:])),"a")
    text_file.write("accuracy {} (% incorrect sites < 75) : {}\n".format(data_name,accuracy))
    std_incorrectly_predicted_sites = incorrectly_predicted_sites_df["Std"].values

    nodes_indexes = np.array(incorrectly_predicted_sites_df.index.str.split("//").str[1]).astype(float).tolist()
    incorrectly_predicted_sites_df = incorrectly_predicted_sites_df.drop(["Average","Std"],axis=1)
    incorrectly_predicted_sites = np.vstack([nodes_indexes,incorrectly_predicted_sites_df.values.T])
    np.save("{}/Incorrectly_Predicted_Sites".format(results_directory), incorrectly_predicted_sites) #Highlight: Hopefully this is the same as before


    fig = plt.figure(figsize=(14, 8))
    ax = fig.add_subplot(111)

    ## the data
    N = average_incorrectly_predicted_sites.shape[0]
    ## necessary variables
    ind = np.arange(N) # the x locations for the groups
    width = 0.35  # the width of the bars
    blue, = sns.color_palette("muted", 1)
    ## the bars
    rects1 = ax.bar(ind, average_incorrectly_predicted_sites,
                    width,
                    color=blue,
                    alpha=0.75,
                    edgecolor="blue",
                    yerr=std_incorrectly_predicted_sites,
                    error_kw=dict(elinewidth=2, ecolor='red'))


    DraupnirUtils.autolabel(rects1,ax)
    # axes and labels
    ax.set_xlim(-width, N* + width)
    ax.set_ylim(0, align_len)
    ax.set_ylabel('Number of incorrect sites')
    ax.set_title('Incorrectly predicted aa sites (%ID); {} \n'.format(data_name) + r'{}'.format(additional_load.full_name))
    xTickMarks = incorrectly_predicted_sites_df.index.tolist()
    ax.set_xticks(ind + width)
    xtickNames = ax.set_xticklabels(xTickMarks)
    plt.setp(xtickNames, rotation=45, fontsize=8)

    plt.savefig("{}/IncorrectlyPredictedAA_BarPlot".format(results_directory))
def build_dataframes_overlapping_histograms(predictions_samples, dataset_train,dataset_test, name, num_samples,children_indexes,results_directory,aa_probs,correspondence_dict=None):
    #TODO: remove????
    print("Building Overlapping Histogram...")
    n_children = len(children_indexes)
    def Percent_ID_OBS_OBS():
        "Generate the  %ID of the OBS TEST sequences against the OBS train"
        percent_id_OBS = dict.fromkeys(dataset_train[:,0,1].tolist(), dict.fromkeys(dataset_test[:,0,1].tolist(),[]))
        for i,trainseq in enumerate(dataset_train):
            seq_TRAIN_obs_letters_i = DraupnirUtils.convert_to_letters(dataset_train[i, 2:,0],aa_probs)
            for j,testseq in enumerate(dataset_test):
                seq_TEST_obs_letters_i = DraupnirUtils.convert_to_letters(dataset_test[j, 2:,0],aa_probs)
                pid =DraupnirUtils.perc_identity_pair_seq(seq_TRAIN_obs_letters_i,seq_TEST_obs_letters_i)
                percent_id_OBS[dataset_train[i,0,1].item()][dataset_test[j,0,1].item()] = pid
        return percent_id_OBS
    def Percent_ID_PRED_OBS():
        "Generate the Average and STD %ID of the sampled sequences against the OBS train"
        #percent_id_PRED = dict.fromkeys(Dataset_train[:,0,1].tolist(), dict.fromkeys(predictions_samples[0,:,1].tolist(),{"Average":[],"Std":[]})) #{"Average":[],"Std":[]}
        percent_id_PRED = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
        for i, seq in enumerate(dataset_train):
            seq_TRAIN_obs_letters_i = DraupnirUtils.convert_to_letters(dataset_train[i, 2:,0],aa_probs)
            for j in range(n_children):#for test node #n_children
                all_sampled_index = predictions_samples[:,j,3:]  # All samples for same test seq
                percent_id_i=[] #all samples for the same test sequence
                for w in range(num_samples):
                    seq_sampled_test = DraupnirUtils.convert_to_letters(all_sampled_index[w],aa_probs)
                    pid = DraupnirUtils.perc_identity_pair_seq(seq_TRAIN_obs_letters_i, seq_sampled_test)
                    percent_id_i.append(pid)
                percent_id_PRED[dataset_train[i,0,1].item()][predictions_samples[0,j,1]]["Average"] = np.mean(np.array(percent_id_i),axis=0)
                percent_id_PRED[dataset_train[i,0,1].item()][predictions_samples[0,j,1]]["Std"] = np.std(np.array(percent_id_i), axis=0)
        return percent_id_PRED


    percent_id_PRED = Percent_ID_PRED_OBS()
    percent_id_OBS = Percent_ID_OBS_OBS()

    fig = plt.figure(figsize=(18, 5))
    ax = fig.add_subplot(111)

    ## Indexes where the bars will go
    ind_lines = np.arange(10,len(dataset_train)*10,10 )
    ind_train = np.arange(5,len(dataset_train)*10,10 )  # the x locations for the train nodes
    width = 10/len(dataset_test) -0.3 # the width of the bars
    ind_test = np.arange(0,10,width + 0.3)  # the x locations for the test nodes, taking into account the space between them?
    start = 0
    blue, = sns.color_palette("muted", 1)
    for train_node in dataset_train[:,0,1]:
        for idx,test_node in enumerate(dataset_test[:,0,1]): #TODO: out of range
            rects2 = ax.bar(ind_test[idx] + start ,
                            percent_id_OBS[train_node.item()][test_node.item()],
                            width,
                            color='green',
                            yerr=[0],
                            error_kw=dict(elinewidth=0.01, ecolor='red'),
                            label="OBSERVED")
            rects1 = ax.bar(ind_test[idx] + start,
                            percent_id_PRED[train_node.item()][test_node.item()]["Average"],
                            width,
                            color="orange",
                            alpha=1,
                            yerr=percent_id_PRED[train_node.item()][test_node.item()]["Std"],
                            error_kw=dict(elinewidth=0.01, ecolor='red'),
                            label="SAMPLED")

        start += 10
    # axes and labels
    ax.set_xlim(-width, np.max(ind_train) +5)
    ax.set_ylim(0, 100)
    ax.set_ylabel('%ID')
    ax.set_title('Overlapping Histogram')
    xTickMarks = ['TrainNode_{}'.format(int(i.item())) for i in dataset_train[:,0,1]]
    ax.set_xticks(ind_train)
    xtickNames = ax.set_xticklabels(xTickMarks)
    plt.setp(xtickNames, rotation=45, fontsize=6)
    plt.vlines(ind_lines,ymin=0,ymax=100)
    ## add a legend--> Not working (inside or outside loop)
    ax.legend((rects1,rects2), ('Sampled',"Observed"),loc=1,bbox_to_anchor=(1.15, 1.1))
    plt.savefig("{}/OverlappingHistogram".format(results_directory))
def barplot_aa_replacement(predictions_samples, dataset_test,full_name, num_samples,nodes_indexes,results_directory,aa_probs,correspondence_dict=None):
    """Plots the amino acid replacements occurred during the evolutionary process. It visualizes the frequency of the amino acids have been
    predicted to replace the true amino acid
    :param predictions_samples  [n_nodes,n_samples,align_len]
    :param dataset_test
    :param full_name: complete name of the data set project
    :param num_samples
    :param nodes_indexes
    :param results_directory
    :param aa_probs
    :param correspondence_dict: contains the correspondence among the nodes indexes and their true names"""
    print("Building aa replacement plots...this might take a while")
    results_directory_path = os.path.dirname(results_directory)
    folder_type =os.path.basename(results_directory).split("_")[0]
    aa_types_dict = DraupnirUtils.aminoacid_names_dict(aa_probs) #apply
    for index,aa in enumerate(list(aa_types_dict.keys())):
        for seq_idx,seq_name in zip(range(dataset_test.shape[0]),nodes_indexes):
            seq_name = int(seq_name.item())
            positions = np.where(dataset_test[seq_idx] == index)[0]
            prediction_for_seq = predictions_samples[:,seq_idx,positions] #predictions for positions where there should be aa == aa_type
            if prediction_for_seq.size !=0:#if it's not empty (for example in the benchmark there are no gaps and then it becomes empty)
                aa_count = np.apply_along_axis(lambda x: np.bincount(x, minlength=aa_probs), axis=0,arr=prediction_for_seq.astype("int64")).T
                aa_count = pd.DataFrame(aa_count,index=positions,columns=list(aa_types_dict.keys()))
                fig = plt.figure(figsize=(16, 12))
                #color_map = matplotlib.colors.ListedColormap(plt.cm.tab20c(np.linspace(0, 1,aa_probs)))
                color_map = matplotlib.colors.ListedColormap(["plum","navy","turquoise","peachpuff","palevioletred","red","darkorange","yellow","lime","green","dodgerblue","blue","purple","magenta","grey","maroon","lightcoral","olive","teal","goldenrod","black"])
                aa_count.plot.bar(stacked=True,colormap=color_map)
                plt.legend( prop={'size': 6},bbox_to_anchor = (1.02, 0.9))
                plt.ylabel("aa count")
                plt.xlabel("True positions of {}".format(aa))
                plt.title("Aa replacement for {} in seq {}; \n".format(aa,correspondence_dict[seq_name]) +r"{}".format(full_name))
                plt.savefig("{}/ReplacementPlots_{}/Replacement_{}_seq_{}.png".format(results_directory_path,folder_type,aa,correspondence_dict[seq_name]))
                plt.clf()
                plt.close(fig)












def index_generator(indexes):
    """Call method to subsample in order.
    model (args, iter_num = index_generator())
    plate(..., next(iter_num))"""
    i=0
    while True:
        yield indexes[i]
        i = (i + 1) % len(indexes)
        return i ##?
def train_batch(svi,patristic_matrix,cladistic_matrix,train_loader,args):
    """Regular batch training without shuffling datatasets
    :param svi: pyro infer engine
    :param cladistic_matrix
    :param patristic_matrix
    :param dataloader train_loader: Pytorch dataloader
    :param namedtuple args
    """
    train_loss = 0.0
    seq_lens = []
    for batch_number, dataset in enumerate(train_loader):
        for batch_name, batch_dataset, batch_patristic, batch_blosum in zip(dataset["batch_name"],
                                                                            dataset["batch_data"],
                                                                            dataset["batch_patristic"],
                                                                            dataset["batch_blosum_weighted"]):
            if args.use_cuda:
                batch_dataset = batch_dataset.cuda()
                batch_blosum = batch_blosum.cuda()
                batch_patristic = batch_patristic.cuda()
            seq_lens += batch_dataset[:, 0, 0].tolist()
            train_loss += svi.step(batch_dataset, batch_patristic, cladistic_matrix, batch_blosum)
            # Normalize loss
            # torch.cuda.reset_max_memory_allocated() #necessary?
    normalizer_train = sum(seq_lens)
    total_epoch_loss_train = train_loss / normalizer_train
    return total_epoch_loss_train


def train(svi,patristic_matrix,cladistic_matrix,train_loader,args):
    """Non batched training
    :param svi: pyro infer engine
    :param cladistic_matrix
    :param patristic_matrix
    :param dataloader train_loader: Pytorch dataloader
    """

    train_loss = 0.0
    seq_lens = []
    for batch_number, dataset in enumerate(train_loader):
            seq_lens += dataset[:,0,0].tolist()
            train_loss += svi.step(dataset,patristic_matrix,cladistic_matrix,None) #None is the clade blosum, it's None because here we do not do clade batching
    # Normalize loss
    normalizer_train = sum(seq_lens)
    total_epoch_loss_train = train_loss #/ normalizer_train
    return total_epoch_loss_train
def train_plating(svi,patristic_matrix,cladistic_matrix,train_loader,args):
    """Plated/subsampling training
    :param svi: pyro infer engine
    :param cladistic_matrix
    :param patristic_matrix
    :param dataloader train_loader: Pytorch dataloader
    :param namedtuple args
    """
    #TODO: make the index_generator() work
    train_loss = 0.0
    seq_lens = []
    for batch_number, dataset in enumerate(train_loader):
            seq_lens += dataset[:,0,0].tolist()
            train_loss += svi.step(dataset,patristic_matrix,cladistic_matrix,index_generator(dataset[:,0,1].tolist())) #None is the clade blosum, it's None because here we do not do clade batching
    # Normalize loss
    normalizer_train = sum(seq_lens)
    total_epoch_loss_train = train_loss #/ normalizer_train
    return total_epoch_loss_train
def train_batch_clade(svi,patristic_matrix,cladistic_matrix,train_loader,args):
    """Batch by clade training
    :param svi: pyro infer engine
    :param cladistic_matrix
    :param patristic_matrix
    :param dataloader train_loader: Pytorch dataloader
    :param namedtuple args
    """
    train_loss = 0.0
    seq_lens = []
    for batch_number, dataset in enumerate(train_loader):
        for clade_name,clade_dataset,clade_patristic,clade_blosum in zip(dataset["clade_name"], dataset["clade_data"],dataset["clade_patristic"],dataset["clade_blosum"]):
            if args.use_cuda:
                clade_dataset = clade_dataset.cuda()
                clade_blosum = clade_blosum.cuda()
                clade_patristic = clade_patristic.cuda() #cannot be used like this, we cannot have a variable size latent space
            seq_lens += clade_dataset[:,0,0].tolist()
            train_loss += svi.step(clade_dataset, patristic_matrix,cladistic_matrix,clade_blosum)
            # Normalize loss
            #torch.cuda.reset_max_memory_allocated() #necessary?
    normalizer_train = sum(seq_lens)
    total_epoch_loss_train = train_loss / normalizer_train
    return total_epoch_loss_train

